import lxml.etree as ET

from amilib.util import AmiLogger, Util
from amilib.xml_lib import NS_MAP, XML_NS, SVG_NS

SVG_SVG = "svg"
SVG_G = "g"
SVG_CIRCLE = "circle"
SVG_LINE = "line"
SVG_POLYLINE = "polyline"
SVG_RECT = "rect"
SVG_PATH = "path"

# path attrs
SVG_d = "d"
SVG_H = "H"
SVG_h = "h"
SVG_C = "C"
SVG_c = "c"
SVG_L = "L"
SVG_l = "l"
SVG_M = "M"
SVG_m = "m"
SVG_Q = "Q"
SVG_q = "q"
SVG_V = "V"
SVG_v = "v"

# coord attrs
X1 = "x1"
X2 = "x2"
Y1 = "y1"
Y2 = "y2"

POINTS = "points"
FILL = "fill"
STROKE = "stroke"
STROKE_WIDTH = "stroke-width"

logger = AmiLogger.create_named_logger(__file__)

class AmiSVG:
    """for actually rendering SVG?"""

    def __init__(self):
        pass

    @classmethod
    def create_SVGElement(cls, tag):
        svg_ns = NS_MAP[SVG_NS]
        return ET.Element(f"{{{svg_ns}}}{tag}")

    @classmethod
    def create_svg(cls):
        svg_elem = cls.create_SVGElement(SVG_SVG)
        return svg_elem

    @classmethod
    def create_circle(cls, xy, r, parent=None,  fill="yellow", stroke="red", stroke_width=1):
        circle_elem = None
        if xy and r:
            circle_elem = cls.create_SVGElement(SVG_CIRCLE)
            circle_elem.attrib["cx"] = str(xy[0])
            circle_elem.attrib["cy"] = str(xy[1])
            circle_elem.attrib["r"] = str(r)
            circle_elem.attrib["fill"] = fill
            circle_elem.attrib["stroke"] = stroke
            circle_elem.attrib["stroke-width"] = str(stroke_width)
        if parent is not None:
            parent.append(circle_elem)
        return circle_elem

    @classmethod
    def create_polyline(cls, xy_array, parent=None, fill="yellow", stroke="red", stroke_width=1, ndec=2):
        """
        creates svg:polyline
        :param xy_array: N*2 array of row-wies x,y coords
        :param parent: if not none, use ase parent element
        :param fill: default 'yellow'
        :param stroke: default red
        :param stroke_width: default 1
        :return: lxml namespaced svg:polyline
        """
        polyline_elem = None
        if xy_array:
            polyline_elem = cls.create_SVGElement(SVG_POLYLINE)
            points = ""
            for i, xy in enumerate(xy_array):
                if i > 0:
                    points += " "
                points += str(round(xy[0], ndec))+","+str(round(xy[1], ndec))
            polyline_elem.attrib[POINTS] = points
            polyline_elem.attrib[FILL] = fill
            polyline_elem.attrib[STROKE] = stroke
            polyline_elem.attrib[STROKE_WIDTH] = str(stroke_width)

        if parent is not None:
            parent.append(polyline_elem)
        return polyline_elem

    @classmethod
    def create_rect(cls, x0y0x1y1, parent=None, fill="gray", stroke="blue", stroke_width=0.3):
        """
        requires coords in  form [x1, y1, x2, y2]
        :param x0y0x1y1: coords in  form [x1, y1, x2, y2]
        :param parent: if not None, appends to this
        """
        logger.debug(f"box {x0y0x1y1}")
        svg_rect = AmiSVG.create_SVGElement(SVG_RECT)
        svg_rect.attrib["fill"] = fill
        svg_rect.attrib["stroke"] = stroke
        svg_rect.attrib["stroke-width"] = str(stroke_width)
        try:
            width = x0y0x1y1[2] - x0y0x1y1[0]
            height = x0y0x1y1[3] - x0y0x1y1[1]
        except Exception as e:
            raise ValueError(f"bbox must be four floats [x0, y0, x1, y1], got {x0y0x1y1}")
        svg_rect.attrib["x"] = str(x0y0x1y1[0])
        svg_rect.attrib["y"] = str(x0y0x1y1[1])
        svg_rect.attrib["width"] = str(width)
        svg_rect.attrib["height"] = str(height)
        if parent is not None:
            parent.append(svg_rect)
        return svg_rect

    @classmethod
    def get_x_y_width_height(cls, svg_rect):
        """requires svg_rect to have canonical form x,y,w,h
        """

        x = Util.get_float(svg_rect.attrib.get("x"))
        y = Util.get_float(svg_rect.attrib.get("y"))
        width = Util.get_float(svg_rect.attrib.get("width"))
        height = Util.get_float(svg_rect.attrib.get("height"))
        return (x, y, width, height)

    @classmethod
    def create_canonical_rect(cls, params):
        """doesn't yet do attributes
        params can be [x0, y0, x1, y1] or [[x0, x1], [y0, y1]]"""
        if type(params) is list:
            if len(list) == 4:
                rect = AmiSVG.create_rect()

    @classmethod
    def create_hline(cls, x0, y0, width):
        svg_line = AmiSVG.create_SVGElement(SVG_LINE)
        svg_line.attrib[X1] = str(x0)
        svg_line.attrib[X2] = str(x0 + width)
        svg_line.attrib[Y1] = str(y0)
        svg_line.attrib[Y2] = str(y0)
        return svg_line

    @classmethod
    def create_vline(cls, x0, y0, height):
        svg_line = AmiSVG.create_SVGElement(SVG_LINE)
        svg_line.attrib[X1] = str(x0)
        svg_line.attrib[X2] = str(x0)
        svg_line.attrib[Y1] = str(y0)
        svg_line.attrib[Y2] = str(y0 + height)
        return svg_line




"""
Tests for NLP
"""
from pathlib import Path
import csv

from pytest import approx

from test.resources import Resources
from test.test_all import AmiAnyTest

from amilib.ami_nlp import AmiNLP



class NLPTest(AmiAnyTest):
    """
    Tests for ami_nlp
    """
    # import nltk, string

    def test_compute_text_similarity_STAT(self):
        """
        ami_nlp,cosine_sim compares strings
        :return:
        """
        ami_nlp = AmiNLP()
        assert ami_nlp.cosine_sim('a little bird', 'a little bird') == approx(1.0)
        assert ami_nlp.cosine_sim('a little bird', 'a little bird chirps') == approx(0.7093, abs=0.001)
        assert ami_nlp.cosine_sim('a little bird', 'a big dog barks') == approx(0, abs=0.001)




import time
from io import StringIO
from pathlib import Path

from selenium import webdriver
from selenium.webdriver import Chrome
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.remote.webelement import WebElement
from webdriver_manager.chrome import ChromeDriverManager
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.common.exceptions import ElementClickInterceptedException

import lxml.etree as ET

URL = "url"
XPATH = "xpath"
OUTFILE = "out_file"



class AmiDriver:
    """
    create and wrap a Chrome headless browser
    Author Ayush Garg, modified Peter Murray-Rust
    """

    def __init__(self, sleep=3):
        """
        creates a Chrome WebDriver instance with specified options and settings.
        """
        options = webdriver.ChromeOptions()
        options.page_load_strategy = "none"
        chrome_path = ChromeDriverManager().install()
        chrome_service = Service(chrome_path)
        self.web_driver = Chrome(options=options, service=chrome_service)
        self.lxml_root_elem = None
        self.sleep = sleep

    def set_sleep(self, sleep):
        if sleep is None or sleep < 1:
            print(f"sleep must be >= 1")
            return
        if sleep > 20:
            print(f"sleep must be <= 20")
            sleep = 20
        self.sleep = sleep

#    class AmiDriver:

    def quit(self):
        """quite the web_driver"""
        print("Quitting the driver...")
        self.web_driver.quit()
        print("DONE")

    def safe_click_element(self, element):
        """
        attempt to click on a web element
        in a safe manner, handling potential exceptions and using different strategies if necessary.

        a web browser. It allows you to navigate to web pages, interact with elements on the page, and
        perform various actions
        :param element: the web element to click on. It can be any
        valid web element object, such as an instance of `WebElement` class in Selenium
        :param sleep: time to wait
        """
        assert type(element) is WebElement, f"should be WebElement found {type(element)}"
        try:
            # Wait for the element to be clickable
            WebDriverWait(self, self.sleep).until(
                EC.element_to_be_clickable((By.XPATH, element.get_attribute("outerHTML")))
            )
            print(f"waiting... {self.sleep}")
            element.click()
        except ElementClickInterceptedException:
            # If the element is not clickable, scroll to it and try again
            self.web_driver.execute_script("arguments[0].scrollIntoView();", element)
            element.click()
        except Exception:
            # If it still doesn't work, use JavaScript to click on the element
            self.web_driver.execute_script("arguments[0].click();", element)

    #    class AmiDriver:

    def get_page_source(self, url):
        """
        returns the page source code.

        :param url: The URL of the webpage you want to retrieve the source code from
        :param sleep: sleep time
        :return: the page source of the website after the driver navigates to the specified URL.
        """
        print(f"Fetching page source from URL: {url}")
        self.web_driver.get(url)
        time.sleep(self.sleep)
        return self.web_driver.page_source

    def click_xpath_list(self, xpath_list):
        if not xpath_list or type(xpath_list) is not list:
            print(f"no xpath_list {xpath_list}")
            return
        print(f"Clicking xpaths...{xpath_list}")
        for xpath in xpath_list:
            print(f"xpath: {xpath}")
            self.click_xpath(xpath)

    #    class AmiDriver:

    def click_xpath(self, xpath):
        """
        find clickable elemnts by xpath and click them
        :param xpath: xpath to click
        :param sleep: wait until click has been executed
        """
        print(f">>>>before click {xpath} => {self.get_lxml_element_count()}")
        # elements = self.get_lxml_root().xpath(xpath)
        elements = self.web_driver.find_elements(
            By.XPATH,
            xpath,
        )
        print(f"click found WebElements {len(elements)}")
        for element in elements:
            self.safe_click_element(element)
            print(f"sleep {self.sleep}")
            time.sleep(self.sleep)  # Wait for the section to expand
        print(f"<<<<element count after = {self.get_lxml_element_count()}")

    def get_lxml_element_count(self):
        elements = self.get_lxml_root_elem().xpath("//*")
        return len(elements)

    #    class AmiDriver:

    def download_expand_save(self, url, xpath_list, html_out, level=99, sleep=3, pretty_print=True):
        """
        Toplevel convenience class

        :param url: to download
        :param xpath_list: ordered list of Xpaths to click
        :param html_out: file to write
        :param level: number of levels to desend (default = 99)
        :param sleep: seconds to sleep between download (default 3)
        """
        if url is None:
            print(f"no url given")
            return
        html_source = self.get_page_source(url)
        if xpath_list is None:
            print(f"no xpath_list specified")
        else:
            self.click_xpath_list(xpath_list[:level])

        if html_out is None:
            print(f"no output html")
            return
        print(f"writing ... {html_out}")

    #    class AmiDriver:

    def write_html(self, html_out, html_elem=None, pretty_print=True, debug=False):
        """
        convenience method to write HTML
        :param out_html: output file
        :param html_elem: elem to write, if none uses driver.root_elem
        :param pretty_print: pretty_print (default True)
        :param debug: writes name of file
        """
        if html_elem is None:
            html_elem = self.get_lxml_root_elem()
        ss = ET.tostring(html_elem, pretty_print=pretty_print)
        if debug:
            print(f"writing {html_out}")

        Path(html_out).parent.mkdir(exist_ok=True, parents=True)
        with open(html_out, 'wb') as f:
            f.write(ss)

    #    class AmiDriver:

    def execute_instruction_dict(self, gloss_dict, keys=None):
        keys = gloss_dict.keys() if not keys else keys
        for key in keys:
            _dict = gloss_dict.get(key)
            if _dict is None:
                print(f"cannot find key {key}")
                continue
            self.download_expand_save(_dict.get(URL), _dict.get(XPATH), _dict.get(OUTFILE))

    #    class AmiDriver:

    def get_lxml_root_elem(self):
        """Convenience method to query the web_driver DOM
        :param xpath: to query the dom
        :return: elements in Dom satisfying xpath (may be empty list)
        """
        if self.lxml_root_elem is None:
            data = self.web_driver.page_source
            doc = ET.parse(StringIO(data), ET.HTMLParser())
            self.lxml_root_elem = doc.xpath("/*")[0]
            print(f"elements in lxml_root: {len(self.lxml_root_elem.xpath('//*'))}")
        return self.lxml_root_elem

    #    class AmiDriver:

    def run_from_dict(self, outfile, dikt, declutter=None, keys=None, debug=True):
        """
        reads doc names from dict and creates HTML

        :param outfile: file to write
        :param control: control dict
        :param declutter: elements to remove (default DECLUTTER_BASIC)
        :param keys: list of control keys (default = all)

        """
        self.execute_instruction_dict(dikt, keys=keys)
        root = self.get_lxml_root_elem()
        self.write_html(outfile, pretty_print=True, debug=debug)
        assert Path(outfile).exists(), f"{outfile} should exist"

    #    class AmiDriver:

    def download_and_save(self, outfile, chap, wg, wg_url):
        ch_url = wg_url + f"chapter/{chap}"
        wg_dict = {
            f"wg{wg}_{chap}":
                {
                    URL: ch_url,
                    XPATH: None,  # no expansiom
                    # OUTFILE: spm_outfile_gatsby2 # dont think this gets written
                },
        }
        keys = wg_dict.keys()
        self.run_from_dict(outfile, wg_dict)
        root_elem = self.lxml_root_elem
        self.quit()
        return root_elem






"""bounding box"""

import math
# this should be in a separate library
from amilib.ami_util import AmiUtil



class BBox:
    """bounding box 2array of 2arrays, based on integers
    """

    X = "x"
    Y = "y"
    WIDTH = "width"
    HEIGHT = "height"

    # indexes for convenience
    MIN = 0
    MAX = 1

    def __init__(self, xy_ranges=None, swap_minmax=False):
        """
        Must have a valid bbox
        Still haven'tb worked out logic of default boxes (must include None's)
        as [0.0],[0.0] is valid
        :param xy_ranges: [[x1, x2], [y1, y2]] will be set to integers
        """
        self.xy_ranges = [[], []]
        self.swap_minmax = swap_minmax
        if xy_ranges is not None:
            self.set_ranges(xy_ranges)

    def copy(self):
        """creqtes new BBox as copy of self
        :return: copy of bbox xy_ranges
        """
        bbox = BBox(self.xy_ranges)
        bbox.swap_minmax = self.swap_minmax

        return bbox

    # @classmethod
    # def create_from_numpy_array(cls, nparray):
    #     coords = nparray.tolist()
    #     bbox = BBox(xy_ranges=[[coords[0], coords[2]], [coords[1], coords[3]]])
    #     return bbox

    @classmethod
    def create_from_xy_w_h(cls, xy, width, height):
        """
        create from xy, width height
        all inputs must be floats
        :param xy: origin a [float, float]
        :param width:
        :param height:
        :return:
        """
        assert type(xy[0]) is float or type(xy[0]) is int, f"found {type(xy[0])}"
        assert type(xy[1]) is float or type(xy[1]) is int, f"found {type(xy[1])}"
        assert type(width) is float or type(width) is int, f"found {type(width)}"
        assert type(height) is float or type(height) is int, f"found {type(height)}"

        try:
            xy_ranges = [[float(xy[0]), float(xy[0]) + float(width)], [float(xy[1]), float(xy[1]) + float(height)]]
        except Exception as e:
            raise ValueError(f"cannot create bbox from {xy},{width},{height}, {e}")
        return BBox(xy_ranges=xy_ranges)

    @classmethod
    def create_from_points(cls, points_list, tolerance=0.001):
        """make bounding box if 4 points can be aligned in a rectangle within tolerance
        choose low/low and high/high ; don't use othe values except as True/False
        :param points_list: reqyires len=4
        :param tolerance: must be aligned within tolerance; default 0.001
        :return: BBox [lowx,  lowy] [highx, highy] => [[lowx, highx], [lowy, highy]] or None
        """
        if not points_list:
            return None
        if not points_list or len(points_list) != 4:
            return None
        high_high = cls._find_low_low_high_high(points_list, 1, tolerance)
        low_low = cls._find_low_low_high_high(points_list, -1, tolerance)
        bbox = None
        if low_low and high_high:
            bbox = BBox(xy_ranges=[
                [low_low[0], high_high[0]],
                [low_low[1], high_high[1]]
            ])
        return bbox

    def set_ranges(self, xy_ranges):
        if xy_ranges is None:
            raise ValueError("no lists given")
        if len(xy_ranges) != 2:
            raise ValueError("must be 2 lists of lists")
        if xy_ranges[0] is not None and len(xy_ranges[0]) == 0:
            xy_ranges[0] = None
        if xy_ranges[0] is not None and len(xy_ranges[0]) != 2:
            raise ValueError(f"range {xy_ranges[0]} must be None or 2-tuple")
        if xy_ranges[1] is not None and len(xy_ranges[1]) == 0:
            xy_ranges[1] = None
        if xy_ranges[1] is not None and len(xy_ranges[1]) != 2:
            raise ValueError(f"range {xy_ranges[1]} must be None or 2-tuple")
        self.set_xrange(xy_ranges[0])
        self.set_yrange(xy_ranges[1])

    def get_ranges(self):
        """gets ranges as [xrange, yrange]"""
        return self.xy_ranges

    def set_xrange(self, rrange):
        self.set_range(0, rrange)

    def get_xrange(self):
        return self.xy_ranges[0]

    def get_width(self):
        """get width
        :return: width or None if x range invalid or not set"""
        if self.get_xrange() is None or len(self.get_xrange()) == 0:
            return None
        assert self.get_xrange() is not None
        assert len(self.get_xrange()) == 2, f"xrange, got {len(self.get_xrange())}"
        return self.get_xrange()[1] - self.get_xrange()[0]

    def set_yrange(self, rrange):
        self.set_range(1, rrange)

    def get_yrange(self):
        return self.xy_ranges[1]

    def get_height(self):
        if self.get_yrange() is None or len(self.get_yrange()) == 0:
            return None
        return self.get_yrange()[1] - self.get_yrange()[0] if len(self.get_yrange()) == 2 else None

    def set_range(self, index, rrange):
        if index != 0 and index != 1:
            raise ValueError(f"bad tuple index {index}")
        if rrange is None:
            self.xy_ranges[index] = None
            return
        val0 = int(rrange[0])
        val1 = int(rrange[1])
        # This may cause problem for vertical text
        # if val1 < val0:
        #     if self.swap_minmax:
        #         val1, val0 = val0, val1
        #     else:
        #         raise ValueError(f"ranges must be increasing {val0} !<= {val1}")
        self.xy_ranges[index] = [val0, val1]
        self.xy_ranges[index] = [val0, val1]

    def __str__(self):
        return str(self.xy_ranges)

    def __repr__(self):
        return str(self.xy_ranges)

    def intersect(self, bbox):
        """
        inclusive intersection of boxes (AND)
        if any fields are empty returns None
        DOES NOT CHANGE self
        :param bbox:
        :return: new Bbox (max(min) ... (min(max)) or None if any  errors
        """
        bbox1 = None
        if bbox is not None:
            xrange = self.intersect_range(self.get_xrange(), bbox.get_xrange())
            yrange = self.intersect_range(self.get_yrange(), bbox.get_yrange())
            bbox1 = BBox([xrange, yrange]) if xrange and yrange else None
            # print(f"self {self} bbox {bbox} => {bbox1}")
        return bbox1

    def union(self, bbox):
        """
        inclusive merging of boxes (OR)
        if any fields are empty returns None
        DOES NOT CHANGE self

        :param bbox:
        :return: new Bbox (min(min) ... (max(max)) or None if any  errors
        """
        bbox1 = None
        if bbox is not None:
            xrange = self.union_range(self.get_xrange(), bbox.get_xrange())
            yrange = self.union_range(self.get_yrange(), bbox.get_yrange())
            bbox1 = BBox((xrange, yrange))
        return bbox1

    @classmethod
    def intersect_range(cls, range0, range1):
        """intersects 2 range tuples"""
        if len(range0) != 2 or len(range1) != 2:
            return None
        min_coord = max(range0[cls.MIN], range1[cls.MIN])
        max_coord = min(range0[cls.MAX], range1[cls.MAX])
        if max_coord < min_coord:
            return None
        return [min_coord, max_coord]

    @classmethod
    def union_range(cls, range0, range1):
        """intersects 2 range tuples"""
        rrange = []
        if len(range0) == 2 and len(range1) == 2:
            rrange = [min(range0[0], range1[0]), max(range0[1], range1[1])]
        return rrange

    def add_coordinate(self, xy_tuple):
        assert xy_tuple is not None, f"xy_tuple must have coordinates"
        self.add_to_range(0, self.get_xrange(), xy_tuple[0])
        self.add_to_range(1, self.get_yrange(), xy_tuple[1])

    def add_to_range(self, index, rrange, coord):
        """if coord outside range , expand range

        :param index: 0 or  =1 for axis
        :param rrange: x or y range
        :param coord: x or y coord
        :return: None (changes range)
        """
        if index != 0 and index != 1:
            raise ValueError(f"bad index {index}")
        if len(rrange) != 2:
            rrange = [None, None]
        if rrange[0] is None or coord < rrange[0]:
            rrange[0] = coord
        if rrange[1] is None or coord > rrange[1]:
            rrange[1] = coord
        self.xy_ranges[index] = rrange
        return rrange

    @classmethod
    def create_box(cls, xy, width, height, use_int=True):
        if xy is None or width is None or height is None:
            raise ValueError("All params must be not None")
        if use_int:
            width = int(width)
            height = int(height)
        if len(xy) != 2:
            raise ValueError("xy must be an array of 2 values")
        if width < 0 or height < 0:
            raise ValueError("width and height must be non negative")
        xrange = [xy[0], xy[0] + width]
        yrange = [xy[1], xy[1] + int(height)]
        bbox = BBox.create_from_ranges(xrange, yrange)
        return bbox

    def expand_by_margin(self, margin):
        """
        if margin is scalar, apply to both axes
        if margin is 2-tuple, apply to x and y separately
        if margin is negative applies only if current range is >- 2*margin + 1
        i
        :param margin: scalar dx or tuple (dx, dy)
        :return: None
        """

        if not isinstance(margin, list):
            margin = [margin, margin]
        self.change_range(0, margin[0])
        self.change_range(1, margin[1])

    def change_range(self, index, margin):
        """
        change range by margin

        :param index: 0 for X 1 for Y
        :param margin:
        :return:
        """
        if index != 0 and index != 1:
            raise ValueError(f"Bad index for range {index}")
        rr = self.xy_ranges[index]
        rr[0] -= margin
        rr[1] += margin
        # range cannot be <= 0
        if rr[0] >= rr[1]:
            mid = (rr[0] + rr[1]) / 2
            rr[0] = int(mid - 0.5)
            rr[1] = int(mid + 0.5)
        self.xy_ranges[index] = rr

    @classmethod
    def create_from_ranges(cls, xr, yr):
        """
        create from 2 2-arrays
        :param xr: 2-list of range
        :param yr: 2-list of range
        :return:
        """
        bbox = BBox()
        bbox.set_xrange(xr)
        bbox.set_yrange(yr)
        return bbox

    def is_valid(self):
        """
        both ranges must be present and non-negative
        :return:
        """
        if self.xy_ranges is None or len(self.xy_ranges) != 2:
            return False
        if self.xy_ranges[0] is None or self.xy_ranges[1] is None:
            return False
        try:
            ok = self.get_width() >= 0 and self.get_height() >= 0
            return ok
        except Exception:
            return False

    def set_invalid(self):
        """set xy_ranges to None"""
        self.xy_ranges = None

    @classmethod
    def get_width_height(cls, bbox):
        """
        TODO MOVED

        :param bbox: tuple of tuples ((x0,x1), (y0,y1))
        :return: (width, height) tuple
        """
        """
        needs to have its own class
        """
        width = bbox[0][1] - bbox[0][0]
        height = bbox[1][1] - bbox[1][0]
        return width, height

    @classmethod
    def fits_within(cls, bbox, bbox_gauge):
        """
        does bbox fit within self (relative coordinates)

        Will this parcel fir the letter box?


        TODO MOVED

        :param bbox: tuple of tuples ((x0,x1), (y0,y1))
        :param bbox_gauge: tuple of (width, height) that bbox must fit in
        :return: true if fits in rectangle
        """
        """
        needs to have its own class
        """
        width, height = bbox.get_width_height()
        return width < bbox_gauge[0] and height < bbox_gauge[1]

    def min_dimension(self):
        """
        gets minimum of height and width
        :return: min(height, width)
        """
        return min(self.get_width(), self.get_height())

    def max_dimension(self):
        """
        gets maximum of height and width
        :return: max(height, width)
        """
        return max(self.get_width(), self.get_height())

    def get_point_pair(self):
        """
        BBox stores the location as ranges of x and y values:
        [[x1, x2], [y1, y2]]
        sometimes it is necessary to work with points instead:
        [(y1, x1), (y2, x2)]
        :returns: list of 2 tuples
        """
        return [(self.get_yrange()[0], self.get_xrange()[0]),
                (self.get_yrange()[1], self.get_xrange()[1])]
        # remember that the indexing is in terms of rows and columns
        # hence x(columns) y(rows) values are flipped when returning point pair

    # @classmethod
    # def plot_bbox_on(cls, image, bbox):
    #     """
    #     Plots bbox on an image
    #     :param: image
    #     :type: numpy array
    #     :param: bbox
    #     :type: BBox or list
    #     :returns: fig, ax
    #     """
    #     from skimage import draw
    #     pixel_value = 200  # 0 is black
    #     # bbox can either be BBox object or in form of [[a, b][c, d]]
    #
    #     # if type(bbox) == BBox:
    #     #     assert bbox.is_valid()
    #     # elif type(bbox) == list:
    #     #     bbox = BBox(bbox)
    #     #     assert bbox.is_valid()
    #     # else:
    #     #     # the bbox passed is not invalid
    #     #     return None
    #
    #     point_pair = bbox.get_point_pair()
    #     if point_pair[0][0] > image.shape[0] or point_pair[0][1] > image.shape[1]:
    #         # if the starting point is outside the image, ignore bbox
    #         return image
    #
    #     try:
    #         row, col = draw.rectangle_perimeter(start=point_pair[0], end=point_pair[1])
    #         image[row, col] = pixel_value
    #     except IndexError:
    #         point_pair = BBox.fit_point_pair_within_image(image, point_pair)
    #         row, col = draw.rectangle_perimeter(start=point_pair[0], end=point_pair[1])
    #         image[row, col] = pixel_value
    #
    #     return image

    @classmethod
    def fit_point_pair_within_image(cls, image, point_pair):
        max_row = image.shape[0]
        max_col = image.shape[1]
        bbox_row = point_pair[1][0]
        bbox_col = point_pair[1][1]
        if bbox_row >= max_row - 1:
            bbox_row = max_row - 2
        if bbox_col >= max_col - 1:
            bbox_col = max_col - 2
        point_pair[1] = (bbox_row, bbox_col)
        return point_pair

    @classmethod
    def create_from_corners(cls, xy1, xy2):
        if xy1 is None or xy2 is None:
            return None
        if len(xy1) != 2 or len(xy2) != 2:
            return None
        xrange = [xy1[0], xy2[0]] if xy2[0] > xy1[0] else [xy2[0], xy1[0]]
        yrange = [xy1[1], xy2[1]] if xy2[1] > xy1[1] else [xy2[1], xy1[1]]
        bbox = BBox(xy_ranges=[xrange, yrange])
        return bbox

    @property
    def centroid(self):
        return [
            (self.get_xrange()[0] + self.get_xrange()[1]) / 2,
            (self.get_yrange()[0] + self.get_yrange()[1]) / 2
        ]

    def contains_point(self, point):
        """does point lie within xy_ranges inclusive
        :param point: 2D numeric array [x, y] or 2-tuple
        :return: False if point is None or self is invalid or point lies outside
        """
        if not BBox.validate_point(point) or not self.is_valid():
            return False
        if point[0] < self.xy_ranges[0][0] or point[0] > self.xy_ranges[0][1]:
            return False
        if point[1] < self.xy_ranges[1][0] or point[1] > self.xy_ranges[1][1]:
            return False
        return True

    def contains_geom_object(self, geom_object):
        """does object fit within self (inclusive coords)

        :param geom_object: 2-float array or BBox
        :return: None returns False
        """

        if not geom_object:
            return False
        if AmiUtil.is_point(geom_object):
            return (self.get_xrange()[0] <= geom_object[0] <= self.get_xrange()[1] and
                    self.get_yrange()[0] <= geom_object[1] <= self.get_yrange()[1])
        if type(geom_object) is BBox:
            if (geom_object.get_xrange()[0] >= self.get_xrange()[0] and
                    geom_object.get_xrange()[1] <= self.get_xrange()[1] and
                    geom_object.get_yrange()[0] >= self.get_yrange()[0] and
                    geom_object.get_yrange()[1] <= self.get_yrange()[1]):
                return True
            return False
        return False

    @classmethod
    def validate_point(cls, point):
        if point is None or len(point) != 2:
            return False
        return AmiUtil.is_number(point[0]) and AmiUtil.is_number(point[1])

    @classmethod
    def _find_low_low_high_high(cls, points_list, sign, tolerance=0.001):
        """
        :param points_list: list of 4 points [x,y]
        :param sign: 1 for high_high or -1 for low_low
        :param tolerance: default 0.001
        :return: point with lowestx, lowesty (sign=-1) or highest, higest (sign=1)
        """
        point = None
        if points_list and len(points_list) == 4:
            for p in points_list:
                if point is None:
                    point = p
                else:
                    if cls._is_movable(p[0], point[0], tolerance, sign) and cls._is_movable(p[1], point[1], tolerance,
                                                                                            sign):
                        point = p
        return point

    @classmethod
    def _is_movable(cls, coord_new, coord_old, tolerance, sign):
        if abs(coord_new - coord_old) < tolerance:
            return True
        if math.copysign(1, coord_new - coord_old) == sign:
            return True

    @classmethod
    def assert_xy_ranges(cls, bbox, target_ranges):
        """asserts that bbox has given xy_ranges (main equality test)
        :param bbox: Bbox
        :param target_ranges: 2-D array of floats [[x0,x1], y0,y1]]
        :except: throws variaous exceptions
        """
        if not bbox:
            if not target_ranges:
                # expected None
                return
            else:
                raise ValueError("parameter/s are None")
        if not bbox.is_valid():
            raise ValueError("Bad bbox {bbox}")
        assert bbox.xy_ranges == target_ranges, f"bbox_xy_ranges {bbox.xy_ranges}, target_ranges {target_ranges}"

    def extract_edges_in_box(self, ami_edges):
        # TODO extend to OR and AND and maybe XOR
        """extract any edges which have at least one node in bbox
        :param ami_edges:
        :return: list of extracted edges
        """

        new_ami_edges = []
        for ami_edge in ami_edges:
            bbox = ami_edge.get_or_create_bbox()
            if bbox.get_width() < 2 or bbox.get_height() < 2:
                continue
            start_node = ami_edge.get_start_ami_node()
            end_node = ami_edge.get_end_ami_node()
            start_xy = start_node.centroid_xy
            end_xy = end_node.centroid_xy
            if (self.contains_geom_object(start_xy) or
                    self.contains_geom_object(end_xy)):
                new_ami_edges.append(ami_edge)
        return new_ami_edges

    @property
    def width(self):
        return self.xy_ranges[0][1] - self.xy_ranges[0][0]

    @property
    def height(self):
        return self.xy_ranges[1][1] - self.xy_ranges[1][0]


"""If you looking for the overlap between two real-valued bounded intervals, then this is quite nice:

def overlap(start1, end1, start2, end2):
    how much does the range (start1, end1) overlap with (start2, end2)
    return max(max((end2-start1), 0) - max((end2-end1), 0) - max((start2-start1), 0), 0)
I couldn't find this online anywhere so I came up with this and I'm posting here."""



"""manages bibligraphy and related stuff"""

import lxml
import lxml.etree
import re
# local
from amilib.util import Util


class Reference:
    SINGLE_REF_REC = re.compile(r"""
                    (?P<pre>.*)   # leading string without bracket
                    \(            # bracket
                    (?P<body>
                    (?:[A-Z]|de|d')
                    .*(?:20|19)\d\d[a-z,]*  # body starts uppcase and ends with date (without brackets)
                    )
                    \)            # trailing bracket
                    (?P<post>.*)  # trailing string without bracket
                    """, re.VERBOSE)
    DOI_REC = re.compile(r".*\s(doi:[^\s]*)\.")  # finds DOI string in running text

    AUTHORS_DATE_REC = re.compile(r"""
    (?P<first>((de )|(d')|(el ))?\s*[A-Z][^\s]+) # doesn't seem to do the prefixes yet
    (?P<others>.+)
    (?P<date>20\d\d[a-z]*)
    """, re.VERBOSE)

    DOI_PROTOCOL = "doi:"
    HTTPS_DOI_ORG = "https://doi.org/"

    def __init__(self):
        self.spans = []

    @classmethod
    def create_ref_from_div(cls, div):
        """create from div which contains one or more spans
        """
        if div is None:
            return None
        ref = Reference()
        ref.div = div
        ref.spans = div.xpath("./span")
        return ref

    def markup_dois_in_spans(self):
        """iterates over contained spans until the doi-containing one is found
        """
        for span in self.spans:
            text = span.text
            doi_match = self.DOI_REC.match(text)
            if doi_match:
                doi_txt = doi_match.group(1)
                if self.DOI_PROTOCOL in doi_txt:
                    doi_txt = doi_txt.replace("doi:https", "https")
                    doi_txt = doi_txt.replace(self.DOI_PROTOCOL, self.HTTPS_DOI_ORG)
                    if doi_txt.startswith(self.DOI_PROTOCOL):
                        doi_txt = "https://" + doi_txt
                    print(f"doi: {doi_txt}")
                    a = lxml.etree.SubElement(span.getparent(), "a")  # to avoid circulkar import of H_A TODO

                    a.attrib["href"] = doi_txt
                    a.text = doi_txt
                    break
            else:
                # print(f"no doi: {text}")
                pass

    @classmethod
    def markup_dois_in_div_spans(cls, ref_divs):
        """creates refs and then marks up the spans
        May be rather specific to IPCC"""
        for div in ref_divs:
            ref = Reference.create_ref_from_div(div)
            spans = div.xpath("./span")
            ref.markup_dois_in_spans()


class Biblioref:
    """in-text pointer to References
    of form:
    Lave 1991
    Lecocq and Shalizi 2014
    Gattuso  et  al.  2018;  Bindoff  et  al.  2019
    IPBES 2019b

    IPCC  2018b:  5.3.1    # fist part only
    """

    def __init__(self):
        self.str = None
        self.first = None
        self.others = None
        self.date = None

    def __str__(self):
        s = f"{self.str} => {self.first}|{self.others}|{self.date}"
        return s

    @classmethod
    def create_refs_from_biblioref_string(cls, brefstr):
        """create from in-text string without the brackets
        :param brefstr: string may contain repeated values
        :return: list of Bibliorefs (may be empty or have one member

        """
        bibliorefs = []
        if brefstr:
            bref = " ".join(brefstr.splitlines()).replace(r"\s+", " ")
            chunks = bref.split(";")
            for chunk in chunks:
                # print(f" chunk {chunk}")
                brefx = Biblioref.create_bref(chunk.strip())
                if brefx:
                    bibliorefs.append(brefx)
        return bibliorefs

    @classmethod
    def create_bref(cls, brefstr):
        """create Biblioref from single string
        :param brefstr: of form 'Author/s date' """

        bref = None
        match = Reference.AUTHORS_DATE_REC.match(brefstr)
        if match:
            bref = Biblioref()
            bref.str = brefstr
            bref.first = match.group("first")
            bref.others = match.group("others")
            bref.date = match.group("date")
        return bref

    @classmethod
    def make_bibliorefs(cls, file):
        chap_elem = lxml.etree.parse(str(file))
        div_spans = chap_elem.xpath(".//div[span]")
        total_bibliorefs = []
        rec = re.compile(Util.SINGLE_BRACKET_RE)
        for div in div_spans:
            for span in div.xpath("./span"):
                match = rec.match(span.text)
                if match:
                    body = match.group('body')
                    bibliorefs = Biblioref.create_refs_from_biblioref_string(body)
                    for biblioref in bibliorefs:
                        total_bibliorefs.append(biblioref)
        return total_bibliorefs


class Publication:
    CHAPTER = "Chapter"
    TECHNICAL_SUMMARY = "Technical Summary"

    @classmethod
    def is_chapter_or_tech_summary(cls, span_text):
        return span_text.startswith(Publication.CHAPTER) or span_text.startswith(Publication.TECHNICAL_SUMMARY)



import csv
from itertools import izip

class AmiCSV:

    def read_transpose_write(self, input, output):
        a = izip(*csv.reader(open("input.csv", "rb")))
        csv.writer(open("output.csv", "wb")).writerows(a)

    def transpose(self, a):

        b = izip(*a)
        return b




# Tests wikipedia and wikidata methods under pytest
import logging
import pprint
import unittest
from pathlib import Path

import requests
from lxml import etree, html

# local
from amilib.wikimedia import WikidataPage, WikidataExtractor, WikidataProperty, WikidataFilter
from amilib.wikimedia import WikidataLookup
from test.resources import Resources
from test.test_all import AmiAnyTest

"""This runs under Pycharm and also
cd pyami # toplevel checkout
python3 -m test.test_wikidata
"""

ami_dictRESOURCES_DIR = Path(Path(__file__).parent.parent, "test", "resources")
# TEMP_DIR = Path(Path(__file__).parent.parent, "temp_oldx_delete")
# DICTIONARY_DIR = Path(os.path.expanduser("~"), "projects", "CEVOpen", "dictionary")
EO_COMPOUND = "eoCompound"
EO_COMPOUND_DIR = Path(Resources.TEST_RESOURCES_DIR, EO_COMPOUND)


# NOTE some of these are lengthy (seconds) as they lookup on the Net

class TestWikidataLookup_WIKI_NET(unittest.TestCase):
    """
    lookup wikidata terms, Ids, Requires NET
    """

    def test_lookup_wikidata_acetone_WIKI_NET(self):
        """
        Lookup single term in Wikidata
        Needs connectivity
        :return:
        """
        term = "acetone"
        wikidata_lookup = WikidataLookup()
        qitem0, desc, wikidata_hits = wikidata_lookup.lookup_wikidata(term)
        assert qitem0 == "Q49546"
        assert desc == "chemical compound"
        # assert wikidata_hits == ['Q49546', 'Q24634417', 'Q329022', 'Q63986955', 'Q4673277']
        assert 'Q49546' in wikidata_hits and len(wikidata_hits) >= 3

    def test_lookup_wiki_properties_chemical_compound_WIKI_NET(self):
        """
        Lookup Wikidata page by Q number and confirm properties
        :return: None
        """
        wiki_page = WikidataPage("Q49546")
        # wiki_page.debug_page()
        description = wiki_page.get_description()
        assert description == "chemical compound"

        type_of_chemical_entity = "Q113145171"  # was "Q11173" (chemical entity)
        qval = wiki_page.get_predicate_object("P31", type_of_chemical_entity)
        assert len(qval) == 1
        wiki_page = WikidataPage("Q24634417")
        qval = wiki_page.get_predicate_object("P31", type_of_chemical_entity)
        assert len(qval) == 0
        # actually a scholarly article
        qval = wiki_page.get_predicate_object("P31", "Q13442814")
        assert len(qval) == 1

    def test_lookup_multiple_terms_solvents_WIKI_NET(self):
        """
        search multiple terms in Wikidata
        """
        terms = ["acetone", "chloroform"]
        wikidata_lookup = WikidataLookup()
        qitems, descs = wikidata_lookup.lookup_items(terms)
        assert qitems == ['Q49546', 'Q172275']
        assert descs == ['chemical compound', 'chemical compound']

    @unittest.skip(reason="No AMI Dict in library")
    def test_lookup_with_ami_dictionary_WIKI_NET_DICT(self):
        """
        Wikidata Lookup of items in AMIDict
        TODO add AmiDict to amilib
        :return:
        """
        terms = [
            "A53T",
            "linkage disequilibrium",
            "transcriptomics"
        ]
        wikidata_lookup = WikidataLookup()
        # qitems, descs = wikidata_lookup.lookup_items(terms)
        temp_dir = Path(AmiAnyTest.TEMP_DIR, "wikidata", "oldx")
        temp_dir.mkdir(exist_ok=True, parents=True)
        dictfile, amidict = AMIDict.create_from_list_of_strings_and_write_to_file(
            terms, title="parkinsons", wikidata=True, directory=temp_dir)
        assert Path(dictfile).exists()

    def test_parse_wikidata_html(self):
        """find Wikidata items with given property
        uses the HTML, tacky but works

        in this case the property is P31 (instance-of) and the value is one of
        three
        <div class="wikibase-snakview-value wikibase-snakview-variation-valuesnak">
                                            <a title="Q11173" href="/wiki/Q11173">chemical compound</a>
                                        </div>

        """
        """
    <div class="wikibase-statementgroupview listview-item" id="P31" data-property-id="P31">
        <div class="wikibase-statementgroupview-property">
            <div class="wikibase-statementgroupview-property-label" dir="auto">
                <a title="Property:P31" href="/wiki/Property:P31">instance of</a>
            </div>
        </div>
        <div class="wikibase-statementlistview">
            <div class="wikibase-statementlistview-listview">
                <div id="Q407418$8A24EA26-7C5E-4494-B40C-65356BBB3AA4" class="wikibase-statementview wikibase-statement-Q407418$8A24EA26-7C5E-4494-B40C-65356BBB3AA4 wb-normal listview-item wikibase-toolbar-item">
                    <div class="wikibase-statementview-rankselector">
                        <div class="wikibase-rankselector ui-state-disabled">
                            <span class="ui-icon ui-icon-rankselector wikibase-rankselector-normal" title="Normal rank"/>
                        </div>
                    </div>
                    <div class="wikibase-statementview-mainsnak-container">
                        <div class="wikibase-statementview-mainsnak" dir="auto">
                            <div class="wikibase-snakview wikibase-snakview-e823b98d1498aa78e139709b1b02f5decd75c887">
                                <div class="wikibase-snakview-property-container">
                                    <div class="wikibase-snakview-property" dir="auto"/>
                                </div>
                                <div class="wikibase-snakview-value-container" dir="auto">
                                    <div class="wikibase-snakview-typeselector"/>
                                    <div class="wikibase-snakview-body">
                                        <div class="wikibase-snakview-value wikibase-snakview-variation-valuesnak">
                                            <a title="Q11173" href="/wiki/Q11173">chemical compound</a>
                                        </div>
                                        ...
    """
        p31 = Path(EO_COMPOUND_DIR, "p31.html")
        tree = etree.parse(str(p31))
        root = tree.getroot()
        child_divs = root.findall("div")
        assert len(child_divs) == 2  # direct children
        child_divs = root.findall(".//div")
        assert len(child_divs) == 109  # all descendants
        snak_views = root.findall(".//div[@class='wikibase-snakview-body']")
        assert len(snak_views) == 6  # snkaviwes (boxes on right)
        # snak_a_views = root.findall(".//div[@class='wikibase-snakview-body']//a[starts-with(@title,'Q')]")
        snak_a_views = root.xpath(".//div[@class='wikibase-snakview-body']//a[starts-with(@title,'Q')]")
        assert len(snak_a_views) == 5  #
        texts = []
        titles = []
        for a in snak_a_views:
            texts.append(a.text)
            titles.append(a.get('title'))
        # assert texts == ['chemical compound',\n 'medication',\n 'p-menthan-3-ol',\n 'menthane monoterpenoids',\n 'LIPID MAPS']
        assert texts == ['chemical compound', 'medication', 'p-menthan-3-ol', 'menthane monoterpenoids', 'LIPID MAPS']
        assert titles == ['Q11173', 'Q12140', 'Q27109870', 'Q66124573', 'Q20968889']

    def test_get_predicate_value(self):
        """tests xpath working of predicate_subject test"""
        tree = html.parse(str(Path(EO_COMPOUND_DIR, "q407418.html")))
        root = tree.getroot()
        hdiv_p31 = root.xpath(".//div[@id='P31']")
        assert len(hdiv_p31) == 1
        qvals = hdiv_p31[0].xpath(".//div[@class='wikibase-snakview-body']//a[starts-with(@title,'Q')]")
        assert len(qvals) == 5

    def test_page_get_values_for_property_id(self):
        """test get all values in triples with given property
        e.g. ?page wdt:P31 ?p31_value
        USEFUL

        """
        page = WikidataPage.create_wikidata_ppage_from_file(Path(EO_COMPOUND_DIR, "q407418.html"))
        qvals = page.get_qitems_for_property_id("P31")
        assert len(qvals) == 5
        assert qvals[0].text == "chemical compound"

    def test_get_property_list(self):
        """gets property list for a page
        """
        page = WikidataPage.create_wikidata_ppage_from_file(Path(EO_COMPOUND_DIR, "q407418.html"))
        property_list = page.get_data_property_list()
        assert len(property_list) == 72
        assert property_list[0].property_name == "instance of"
        assert property_list[0].id == "P31"

    def test_get_predicate_value_1(self):
        tree = html.parse(str(Path(EO_COMPOUND_DIR, "q407418.html")))
        root = tree.getroot()
        qvals = root.xpath(".//div[@id='P31']")[0].xpath(".//div[@class='wikibase-snakview-body']//a[@title='Q11173']")
        assert len(qvals) == 1
        assert qvals[0].text == 'chemical compound'
        qvals = root.xpath(".//div[@id='P31']//div[@class='wikibase-snakview-body']//a[@title='Q11173']")
        assert len(qvals) == 1
        assert qvals[0].text == 'chemical compound'

    def test_get_wikidata_predicate_value(self):
        """searches for instance-of (P31) chemical_compound (Q11173) in a wikidata page
        TODO allow for reading local files directly
        """
        pred_id = "P31"
        obj_id = "Q11173"
        file = str(Path(EO_COMPOUND_DIR, "q407418.html"))
        page = WikidataPage.create_wikidata_ppage_from_file(file)
        assert page is not None
        qval = page.get_predicate_object(pred_id, obj_id)
        assert qval[0].text == 'chemical compound'

    def test_get_title_of_page(self):
        title = WikidataPage("q407418").get_title()
        assert title == "L-menthol"

    def test_get_alias_list(self):
        aliases = WikidataPage("q407418").get_aliases()
        assert len(aliases) >= 5 and "l-menthol" in aliases

    def test_get_description(self):
        desc = WikidataPage("q407418").get_description()
        # assert desc == "chemical compound"
        assert "compound" in desc  # wikidata changed this!! 'organic compound used as flavouring and for analgesic properties'

    def test_attval_contains(self):
        """does a concatenated attavle contain a word
        <th scope="col" class="wikibase-entitytermsforlanguagelistview-cell wikibase-entitytermsforlanguagelistview-language">Language</th>

        """
        language_elems = WikidataPage("q407418").get_elements_for_attval_containing_word(
            "class",
            "wikibase-entitytermsforlanguagelistview-language")
        assert len(language_elems) == 1
        assert language_elems[0].text == 'Language'

    @unittest.skip("bug is comparison of sets, needs fixing")
    def test_find_left_properties_and_statements(self):
        """
        TODO comparison of retrieved properties
            <div class="wikibase-snaklistview">
                <div class="wikibase-snaklistview-listview">
                    <div class="wikibase-snakview wikibase-snakview-755d14b02a41025911e80439cb6ed31dcc966768">
                        <div class="wikibase-snakview-property-container">
                            <div class="wikibase-snakview-property" dir="auto">
                               <a title="Property:P662" href="/wiki/Property:P662">PubChem CID</a>
                           </div>
                           ...
        """
        # property_list = WikidataPage("q407418").root.xpath(".//"
        #                                                    "div[@class='wikibase-snaklistview']/"
        #                                                    "div[@class='wikibase-snaklistview-listview']/"
        #                                                    "div/"
        #                                                    "div[@class='wikibase-snakview-property-container']/"
        #                                                    "div[@class='wikibase-snakview-property']/"
        #                                                    "a[starts-with(@title,'Property:')]")

        """<div class="wikibase-statementgroupview-property-label" dir="auto">
              <a title="Property:P274" href="/wiki/Property:P274">chemical formula</a></div>
        """
        lookup = "sroperty"
        lookup = "statement"
        property_list = []
        if lookup == "property":
            classx = "wikibase-statementgroupview-property-label"
            selector = f"@class='{classx}'"
            property_selector = f".//div[@class='{classx}']//a[starts-with(@title,'Property:')]"
            property_list = WikidataPage("q407418").root.xpath(property_selector)
        if lookup == "statement":
            """wikibase-statementgroupview listview-item"""
            # selector = f".//div[@class='wikibase-statementgroupview listview-item']"
            selector = f".//div[starts-with(@class,'wikibase-statementgroupview')]"
            selector = f".//div[@data-property-id]"
            selector = f".//div[@data-property-id]/div[@class='wikibase-statementlistview']//a"
            selector = f".//div[@data-property-id]"
            # selector = f".//div[contains(@class,'wikibase-statementgroupview') and contains(@class,'listview-item')]"
            # selector = f".//div[contains(@class,'listview-item')]"
        print(f" selector {selector} ")

        property_list = WikidataPage("q407418").root.xpath(selector)

        # <!-- property-[statement-list] container TOP LEVEL -->
        # <div class="wikibase-statementgroupview listview-item" id="P31" data-property-id="P31">
        # <!-- property-subject container -->
        # <div class="wikibase-statementgroupview-property">
        #   <div class="wikibase-statementgroupview-property-label" dir="auto"><a title="Property:P31" href="/wiki/Property:P31">instance of</a></div>
        # </div>
        # <!-- statementlist container -->
        # <div class="wikibase-statementlistview">
        # <div class="wikibase-statementlistview-listview">
        # <div id="Q407418$8A24EA26-7C5E-4494-B40C-65356BBB3AA4" class="wikibase-statementview wikibase-statement-Q407418$8A24EA26-7C5E-4494-B40C-65356BBB3AA4 wb-normal listview-item wikibase-toolbar-item">
        # <div class="wikibase-statementview-rankselector"><div class="wikibase-rankselector ui-state-disabled">
        # <!-- "button?" -->
        # <span class="ui-icon ui-icon-rankselector wikibase-rankselector-normal" title="Normal rank"></span>
        # </div></div>
        # <div class="wikibase-statementview-mainsnak-container">
        # <div class="wikibase-statementview-mainsnak" dir="auto"><div class="wikibase-snakview wikibase-snakview-e823b98d1498aa78e139709b1b02f5decd75c887">
        # <div class="wikibase-snakview-property-container">
        # <div class="wikibase-snakview-property" dir="auto"></div>
        # </div>
        # <!-- object-value-container -->
        # <div class="wikibase-snakview-value-container" dir="auto">
        # <div class="wikibase-snakview-typeselector"></div>
        # <div class="wikibase-snakview-body">
        # <div class="wikibase-snakview-value wikibase-snakview-variation-valuesnak"><a title="Q11173" href="/wiki/Q11173">chemical compound</a></div>
        # <div class="wikibase-snakview-indicators"></div>
        # </div>
        # </div>
        # </div></div>
        # <div class="wikibase-statementview-qualifiers"></div>
        # </div>
        # <span class="wikibase-toolbar-container wikibase-edittoolbar-container"><span class="wikibase-toolbar wikibase-toolbar-item wikibase-toolbar-container"><span class="wikibase-toolbarbutton wikibase-toolbar-item wikibase-toolbar-button wikibase-toolbar-button-edit"><a href="#" title=""><span class="wb-icon"></span>edit</a></span></span></span>
        # <div class="wikibase-statementview-references-container">
        # <div class="wikibase-statementview-references-heading"><a class="ui-toggler ui-toggler-toggle ui-state-default"><span class="ui-toggler-icon ui-icon ui-icon-triangle-1-s"></span><span class="ui-toggler-label">0 references</span></a><div class="wikibase-tainted-references-container" data-v-app=""><div class="wb-tr-app"><!----></div></div></div>
        # <div class="wikibase-statementview-references "><div class="wikibase-addtoolbar wikibase-toolbar-item wikibase-toolbar wikibase-addtoolbar-container wikibase-toolbar-container"><span class="wikibase-toolbarbutton wikibase-toolbar-item wikibase-toolbar-button wikibase-toolbar-button-add"><a href="#" title=""><span class="wb-icon"></span>add reference</a></span></div></div>
        # </div>
        # </div><div id="Q407418$A25EE807-DBE3-47CA-9272-00BF975DAEA8" class="wikibase-statementview wikibase-statement-Q407418$A25EE807-DBE3-47CA-9272-00BF975DAEA8 wb-normal listview-item wikibase-toolbar-item">
        # <div class="wikibase-statementview-rankselector"><div class="wikibase-rankselector ui-state-disabled">
        # <span class="ui-icon ui-icon-rankselector wikibase-rankselector-normal" title="Normal rank"></span>
        # </div></div>
        # <div class="wikibase-statementview-mainsnak-container">
        # <div class="wikibase-statementview-mainsnak" dir="auto"><div class="wikibase-snakview wikibase-snakview-87cdd435c7bb91eadb3355615e99ee224aa44984">
        # <div class="wikibase-snakview-property-container">
        # <div class="wikibase-snakview-property" dir="auto"></div>
        # </div>
        # <!-- object-value-container -->
        # <div class="wikibase-snakview-value-container" dir="auto">
        # <div class="wikibase-snakview-typeselector"></div>
        # <div class="wikibase-snakview-body">
        # <div class="wikibase-snakview-value wikibase-snakview-variation-valuesnak"><a title="Q12140" href="/wiki/Q12140">medication</a></div>
        # <div class="wikibase-snakview-indicators"></div>
        # <!-- ..... -->
        # </div>
        # </div></div>
        # </div></div>
        # </div></div><div class="wikibase-addtoolbar wikibase-toolbar-item wikibase-toolbar wikibase-addtoolbar-container wikibase-toolbar-container"><span class="wikibase-toolbarbutton wikibase-toolbar-item wikibase-toolbar-button wikibase-toolbar-button-add"><a href="#" title=""><span class="wb-icon"></span>add reference</a></span></div></div>
        # </div>
        # </div>
        # </div>
        # <span class="wikibase-toolbar-container"></span>
        # <span class="wikibase-toolbar-wrapper"><div class="wikibase-addtoolbar wikibase-toolbar-item wikibase-toolbar wikibase-addtoolbar-container wikibase-toolbar-container"><span class="wikibase-toolbarbutton wikibase-toolbar-item wikibase-toolbar-button wikibase-toolbar-button-add"><a href="#" title="Add a new value"><span class="wb-icon"></span>add value</a></span></div></span></div>
        # </div>"""
        # """wikibase-statementgroupview listview-item"""

        wikidata_page = WikidataPage("q407418")
        data_property_list = wikidata_page.get_data_property_list()
        print(f" data properties {data_property_list}")
        property_set = set(data_property_list)
        print(f"set {property_set}")
        assert 100 >= len(property_set) >= 70
        expected = set([
            'P31', 'P279', 'P361', 'P117', 'P8224', 'P2067', 'P274', 'P233', 'P2017', 'P2054'])
        difference = expected.symmetric_difference(property_set)
        print(f"diff {len(difference)} {difference}")
        assert expected.issubset(property_set), f"not found in {property_set}"
        assert set(wikidata_page.get_property_id_list()[:10]).difference(expected) == set()
        assert wikidata_page.get_property_name_list()[:10] == [
            'instance of', 'subclass of', 'part of', 'chemical structure', 'molecular model or crystal lattice model',
            'mass', 'chemical formula', 'canonical SMILES', 'isomeric SMILES', 'density']
        property_list = wikidata_page.get_data_property_list()
        assert 108 >= len(property_list) >= 70
        # assert wikidata_page.get_property_id_list()[:10] == [
        #     'P31', 'P279', 'P361', 'P117', 'P8224', 'P2067', 'P274', 'P233', 'P2017', 'P2054']
        assert 'P31' in wikidata_page.get_property_id_list()
        # assert wikidata_page.get_property_name_list()[:10] == [
        #     'instance of', 'subclass of', 'part of', 'chemical structure', 'molecular model or crystal lattice model',
        #     'mass', 'chemical formula', 'canonical SMILES', 'isomeric SMILES', 'density']
        assert "instance of" in wikidata_page.get_property_name_list()

        properties_dict = WikidataProperty.get_properties_dict(property_list)
        dict_str = pprint.pformat(properties_dict)
        print(f"\ndict: \n"
              f"{dict_str}")
        assert properties_dict['P662'] == {'name': 'PubChem CID', 'value': '16666'}

    # all wikidata asserts are fragile
    # assert properties_dict['P31'] == {'name': 'instance of',
    #                                   'statements': {
    #                                       'Q11173': 'chemical compound',
    #                                       'Q12140': 'medication',
    #                                       'Q27109870': 'p-menthan-3-ol',
    #                                       'Q66124573': 'menthane monoterpenoids'}}

    # retains the order of addition
    # skip, fragile
    # assert sorted(list(properties_dict.keys())) == sorted([
    #     'P31', 'P279', 'P361', 'P117', 'P8224', 'P2067', 'P274', 'P233', 'P2017', 'P2054', 'P2101', 'P2102',
    #     'P2128', 'P2275', 'P703', 'P2175', 'P129', 'P366', 'P2789', 'P2868', 'P1343', 'P1987', 'P1748', 'P527',
    #     'P1889', 'P935', 'P373', 'P910', 'P268', 'P227', 'P8189', 'P244', 'P234', 'P235', 'P231', 'P661',
    #     'P662', 'P1579', 'P683', 'P592', 'P6689', 'P4964', 'P679', 'P2062', 'P3117', 'P8494', 'P2840', 'P232',
    #     'P2566', 'P5930', 'P8121', 'P7049', 'P595', 'P3345', 'P715', 'P2057', 'P2064', 'P652', 'P2115', 'P665',
    #     'P2063', 'P8313', 'P1417', 'P646', 'P1296', 'P8408', 'P6366', 'P2004', 'P10283', 'P3417', 'P5076', 'P5082'
    # ])

    #     @unittest.skip("dictionary not included")
    #     def test_update_dictionary_with_wikidata_ids(self):
    #         """Update dictionary by adding Wikidata IDs where missing"""
    #         """
    # <dictionary title="dict_5">
    #     <entry name="allyl isovalerate" term="allyl isovalerate"></entry>
    #     <entry name="allyl octanoate" term="allyl octanoate" wikidataID="Q27251951"></entry>
    #     <entry name="allylhexanoate" term="allylhexanoate" wikidataID="Q3270746"></entry>
    #     <entry name="alpha-alaskene" term="alpha-alaskene"></entry>  <!-- not in Wikidata -->
    #     <entry name="alpha-amyrenone" term="alpha-amyrenone"></entry> <!-- not in Wikidata -->
    # </dictionary>        """
    #         path = Path(EO_COMPOUND_DIR, "dict_5.xml")
    #         # dictionary = AmiDictionary.create_from_xml_file(str(path))
    #         dictionary = None
    #         assert len(dictionary.entries) == 5
    #         entry = dictionary.get_lxml_entry("allylhexanoate")
    #         assert entry.get(WIKIDATA_ID) == "Q3270746"
    #
    #         entry = dictionary.get_lxml_entry("allyl isovalerate")
    #         assert entry.get(WIKIDATA_ID) is None
    #         dictionary.lookup_and_add_wikidata_to_entry(entry)
    #         assert entry.get(WIKIDATA_ID) == "Q27155908"
    #
    #         dictionary.write_to_file(Path(AmiAnyTest.TEMP_DIR, EO_COMPOUND, "dict_5.xml"))
    #
    #     @unittest.skip("LONG DOWNLOAD")
    #     def test_add_wikidata_to_complete_dictionary_with_filter(self):
    #         """Takes existing dictionary and looks up Wikidata stuff for entries w/o WikidataID
    #         Need dictionary in AmiDictionary format"""
    #         input_dir = EO_COMPOUND_DIR
    #         output_dir = Path(AmiAnyTest.TEMP_DIR, EO_COMPOUND)
    #         start_entry = 0
    #         end_entry = 10
    #         input_path = Path(input_dir, "eoCompound1.xml")
    #         assert input_path.exists(), f"{input_path} should exist"
    #         dictionary = AmiDictionary.create_from_xml_file(str(input_path))
    #         assert len(dictionary.entries) == 2114
    #         description = "chemical compound"
    #
    #         for entry in dictionary.entries[start_entry: end_entry]:
    #             wikidata_id = AmiDictionary.get_wikidata_id(entry)
    #             if not AmiDictionary.is_valid_wikidata_id(wikidata_id):
    #                 term = AmiDictionary.get_term(entry)
    #                 # print(f"no wikidataID in entry: {term}")
    #                 dictionary.lookup_and_add_wikidata_to_entry(entry, allowed_descriptions=description)
    #                 wikidata_id = AmiDictionary.get_wikidata_id(entry)
    #                 if wikidata_id is None:
    #                     print(f"no wikidata entry for {term}")
    #                     entry.attrib[WIKIDATA_ID] = AmiDictionary.NOT_FOUND
    #                 else:
    #                     print(f"found {wikidata_id} for {term} desc = {entry.get('desc')}")
    #         dictionary.write_to_file(Path(output_dir, "eoCompound1.xml"))
    #
    #     @unittest.skip("dictionary not included")
    #     def test_disambiguation(self):
    #         """attempts to disambiguate the result of PMR-wikidata lookup
    #         """
    #         input_dir = EO_COMPOUND_DIR
    #         output_dir = Path(AmiAnyTest.TEMP_DIR, EO_COMPOUND)
    #         output_dir.mkdir(exist_ok=True)
    #         input_path = Path(input_dir, "disambig.xml")
    #         assert input_path.exists(), f"{input_path} should exist"
    #         # Dictionary not available yet
    #         # dictionary = AmiDictionary.create_from_xml_file(str(input_path))
    #         dictionary = None
    #         assert len(dictionary.entries) == 9
    #         allowed_descriptions = {
    #             "chemical compound": "0.9",
    #             "group of isomers": "0.7",
    #         }
    #         dictionary = AmiDictionary.create_from_xml_file(str(input_path))
    #         for entry in dictionary.entries:
    #             dictionary.disambiguate_wikidata_by_desc(entry)
    #         dictionary.write_to_file(Path(output_dir, "disambig.xml"))
    #
    #     @unittest.skip("Dictionary not included")
    #     def test_extract_multiple_wikidata_hits_gwp(self):
    #         """
    #         test multiple hits for 'GHG' and use heuristics to find the most likely
    #         uses dictionary created from abbreviations via docanalysis
    #         requires WIKIDATA LOOKUP on Internet
    #
    #         gwp entry is SIMILAR to:
    #   <entry term="GHG" name="Greenhouse Gas" >
    #     <raw wikidataID="['Q167336', 'Q3588927', 'Q925312', 'Q57584895', 'Q110612403', 'Q112192791', 'Q140182']"/>
    #   </entry>
    #
    #         """
    #
    #         ami_dict = AmiDictionary.create_from_xml_file(Resources.TEST_IPCC_CHAP02_ABB_DICT)
    #         ami_dict = None
    #         assert ami_dict is not None
    #         gwp = "GWP"
    #         gwp_element = ami_dict.get_lxml_entry(gwp)
    #         assert gwp_element, f"entry for {gwp} is None, probably not found"
    #         assert type(gwp_element) is lxml.etree._Element, f"entry has type {type(gwp_element)}"
    #         gwp_entry = AmiEntry.create_from_element(gwp_element)
    #         assert type(gwp_entry) is AmiEntry, f"ami_entry is type {type(gwp_entry)}"
    #         assert gwp_entry is not None
    #         wikidata_id = gwp_entry.wikidata_id
    #         assert not wikidata_id
    #         raw_wikidata_ids = gwp_entry.get_raw_child_wikidata_ids()
    #         assert len(raw_wikidata_ids) == 6
    #         assert set(raw_wikidata_ids) == {'Q901028', 'Q57084968', 'Q57402965', 'Q57084921', 'Q57084755', 'Q57084776'}
    #
    #     @unittest.skip("dictiomary not included")
    #     def test_get_best_match_for_gwp(self):
    #         """
    #         gets best wikidata match for term
    #         WORKING
    # entry like:
    #   <entry term="GHG" name="Greenhouse Gas" >
    #     <raw wikidataID="['Q167336', 'Q3588927', 'Q925312', 'Q57584895', 'Q110612403', 'Q112192791', 'Q140182']"/>
    #   </entry>
    #
    #         """
    #         ami_dict = AmiDictionary.create_from_xml_file(Resources.TEST_IPCC_CHAP02_ABB_DICT)
    #         ami_dict = None
    #         term = "GWP"
    #         ami_entry = ami_dict.get_ami_entry(term)
    #         matched_pages = ami_entry.get_wikidata_pages_from_raw_wikidata_ids_matching_wikidata_page_title()
    #         assert len(matched_pages) == 1 and matched_pages[0].get_id() == "Q901028"

    def test_get_instances(self):
        """<div class="wikibase-statementview-mainsnak-container">
<div class="wikibase-statementview-mainsnak" dir="auto"><div class="wikibase-snakview wikibase-snakview-e823b98d1498aa78e139709b1b02f5decd75c887">
<div class="wikibase-snakview-property-container">
<div class="wikibase-snakview-property" dir="auto"></div>
</div>
<div class="wikibase-snakview-value-container" dir="auto">
<div class="wikibase-snakview-typeselector"></div>
<div class="wikibase-snakview-body">
<div class="wikibase-snakview-value wikibase-snakview-variation-valuesnak"><a title="Q11173" href="/wiki/Q11173">chemical compound</a></div>
<div class="wikibase-snakview-indicators"></div>
</div>
</div>
</div></div>
<div class="wikibase-statementview-qualifiers"></div>
</div>"""
        pass

    # @unittest.skip("LONG DOWNLOAD")
    # def test_add_wikidata_to_imageanalysis_output(self):
    #     """creates dictionary from list of terms and looks up Wikidata"""
    #     terms = [
    #         "isopentyl-diphosphate delta-isomerase"
    #         "squalene synthase",
    #         "squalene monoxygenase",
    #         "phytoene synthase",
    #         "EC 2.5.1.6",
    #         "EC 4.4.1.14",
    #         "EC 1.14.17.4",
    #         "ETRL",
    #         "ETR2",
    #         "ERS1",
    #         "EIN4",
    #     ]
    #     with open(Path(RESOURCES_DIR, EO_COMPOUND, "compounds.txt"), "r") as f:
    #         terms = f.readlines()
    #         assert 100 > len(terms) > 87
    #     wikidata_lookup = WikidataLookup()
    #     # qitems, descs = wikidata_lookup.lookup_items(terms)
    #     temp_dir = Path(TEMP_DIR, "wikidata")
    #     temp_dir.mkdir(exist_ok=True)
    #     # limit = 10000
    #     limit = 5
    #     amidict, dictfile = AmiDictionary.create_dictionary_from_words(terms[:limit], title="compounds",
    #                                                                               wikidata=True, outdir=temp_dir)
    #     print(f"wrote to {dictfile}")
    #     assert os.path.exists(dictfile)

    def test_wikidata_extractor(self):
        query = '2-fluorobenzoic acid'
        extractor = WikidataExtractor('en')
        id = extractor.search(query)
        id_dict = extractor.load(id)
        print(id_dict)

    def test_simple_wikidata_query(self):
        """get ID list for query results
        see https://www.wikidata.org/w/api.php for options
        each entry has a small number of attributes (e.g. description, URL, )"""
        query = "isomerase"
        url_str = f"https://www.wikidata.org/w/api.php?" \
                  f"action=wbsearchentities" \
                  f"&search={query}" \
                  f"&language=en" \
                  f"&format=json"
        response = requests.get(url_str)
        js = response.json()
        # print(pprint.pformat(js))

    def test_wikidata_id_lookup(self):
        """test query wikidata by ID
        """
        ids = "Q11966262"  # "Dyschirius politus" a species of insect
        url_str = f"https://www.wikidata.org/w/api.php?" \
                  f"action=wbgetentities" \
                  f"&ids={ids}" \
                  f"&language=en" \
                  f"&format=json"
        response = requests.get(url_str)
        response_js = response.json()["entities"][ids]
        # print(f"pages for {ids}\n", pprint.pformat(response_js))
        assert list(response_js.keys()) == ['pageid', 'ns', 'title', 'lastrevid', 'modified', 'type', 'id', 'labels',
                                            'descriptions', 'aliases', 'claims', 'sitelinks']
        assert response_js["id"] == ids
        assert response_js["title"] == "Q11966262"
        assert response_js["labels"]["en"]["value"] == "Dyschirius politus"
        assert response_js["descriptions"]["en"]["value"] == "species of insect"
        key_set = set(list(response_js["claims"].keys()))
        test_set = set(
            list(['P225', 'P105', 'P171', 'P31', 'P685', 'P846', 'P1939', 'P373', 'P815', 'P3151', 'P3186',
                  'P3405', 'P2464', 'P1843', 'P7202', 'P7552', 'P6105', 'P6864', 'P8915', 'P3240', 'P2671',
                  'P3606', 'P8707', 'P10243'])
        )
        assert test_set <= key_set

        ids = "P117"  # chemical compound
        url_str = f"https://www.wikidata.org/w/api.php?" \
                  f"action=wbgetentities" \
                  f"&ids={ids}" \
                  f"&language=en" \
                  f"&format=json"
        response = requests.get(url_str)
        response_js = response.json()["entities"][ids]
        # print(f"pages for {ids}\n", pprint.pformat(response_js))
        assert list(response_js.keys()) == ['pageid', 'ns', 'title', 'lastrevid', 'modified', 'type', 'datatype', 'id',
                                            'labels', 'descriptions', 'aliases', 'claims']
        assert response_js["id"] == "P117"
        assert response_js["title"] == "Property:P117"
        assert response_js["labels"]["en"]["value"] == "chemical structure"
        assert response_js["descriptions"]["en"][
                   "value"] == "image of a representation of the structure for a chemical compound"
        assert set(['P31', 'P1855', 'P3254', 'P2302', 'P1629', 'P1647', 'P2875', 'P1659']).issubset(
            set(list(response_js["claims"].keys()))
        )

    #        wikidata_page = WikidataPage.create_from_response(response)

    def test_multiple_ids(self):
        ids = "P31|P117"
        url_str = f"https://www.wikidata.org/w/api.php?action=wbgetentities&ids={ids}&language=en&format=json"
        response = requests.get(url_str)
        json_dict = response.json()
        assert list(json_dict['entities'].keys()) == ['P31', 'P117']
        assert list(json_dict['entities']['P117'].keys()) == [
            'pageid', 'ns', 'title', 'lastrevid', 'modified', 'type', 'datatype', 'id', 'labels',
            'descriptions', 'aliases', 'claims']
        assert json_dict['entities']['P117']['labels']['en']['value'] == 'chemical structure'
        assert json_dict['entities']['P31']['labels']['en']['value'] == 'instance of'

    def test_read_wikidata_filter(self):
        path = Path(Resources.TEST_RESOURCES_DIR, "filter00.json")
        assert path.exists(), f"{path} should exist"
        filter = WikidataFilter.create_filter(path)
        assert filter.json['plugh'] == "xyzzy"
        assert filter.json['filter']['description'] == "chemical"
        assert filter.json['filter'][
                   'regex'] == "(chemical compound|chemical element)", f"found {filter.json['filter']['regex']}"


class WikimediaTests:
    @classmethod
    def test_sparql_wrapper_WIKI(cls):
        """A
        uthor Shweata M Hegde
        from wikidata query site
        """

        query = """#research council
        SELECT ?researchcouncil ?researchcouncilLabel 
        WHERE 
        {
          ?researchcouncil wdt:P31 wd:Q10498148.
          SERVICE wikibase:label_xml { bd:serviceParam wikibase:language "[AUTO_LANGUAGE],en". }
        }"""

        results = WS.get_results_xml(query)
        print(results)




if __name__ == '__main__':
    unittest.main()



import argparse
import logging
import re
import sys
import textwrap
from enum import Enum
from pathlib import Path

from amilib.ami_html import HTMLArgs
from amilib.pdf_args import PDFArgs
from amilib.util import AbstractArgs
from amilib.wikimedia import WikidataLookup

# local

AMIX_DIR = Path(__file__).parent
REPO_DIR = AMIX_DIR.parent


class AmiLib:
    logger = logging.getLogger("amilib")
    #     """ main entry point for running pyami
    #      """
    #     OUTFILE = "outfile"
    #
    #     # flags
    #     APPLY = "apply"
    #     ASSERT = "assert"
    #     CHECK_URLS = "check_urls"
    COPY = "copy"
    #     COMBINE = "combine"
    #     CONFIG = "config"
    #     CONTAINS = "contains"
    DEBUG = "debug"
    #     DELETE = "delete"
    #     DEST = "dest"
    #     DICTIONARY = "dictionary"
    EXAMPLES = "examples"
    #     FILTER = "filter"
    FLAGS = "flags"
    GLOB = "glob"
    #     LOOKUP = "lookup"
    #     PRINT_SYMBOLS = "print_symbols"
    PROJ = "proj"
    #     RECURSE = "recurse"
    #     REGEX = "regex"
    SECT = "sect"
    #     SRC = "src"
    SPLIT = "split"
    #     SYMBOLS = "symbols"
    SYSTEM_EXIT_OK = "SystemExitOK"
    SYSTEM_EXIT_FAIL = "SystemExitFail_"
    #     TEST = "test"
    VERSION = "version"
    #     WIKIDATA_SPARQL = "wikidata_sparql"
    #     XPATH = "xpath"
    #     # apply methods 1:1 input-output
    #     # obsolete (use Converters)
    #     PDF2TXT = "pdf2svg"
    #     PDF2SVG = "pdf2svg"
    #     SVG2PAGE = "svg2page"
    #     XML2HTML = "xml2html"
    #     TXT2SENT = "txt2sent"
    #     XML2TXT = "xml2txt"
    #     TXT2PARA = "txt2para"
    #     XML2SECT = "xml2sect"
    #     # combine methods n:1 input-output
    #     CONCAT_STR = "concat_str"
    #     # split methods 1:n input-output
    #     # symbols to update table
    #     SPECIAL_SYMBOLS = ["_proj"]
    LOGLEVEL = "loglevel"

    #     PY4AMI = "pyamihtmlx"
    #
    #     # parsers are these used??
    #     # DICT_PARSER = "DICT"
    #     # HTML_PARSER = "HTML"
    #     # IPCC_PARSER = "IPCC"
    #     # PDF_PARSER  = "PDF"
    #     # PROJECT_PARSER = "PROJECT"
    #     # UNFCCC_PARSER  = "UNFCCC"
    #
    #     logger = logging.getLogger("pyami")
    #     symbol_ini = None
    #
    def __init__(self):
        """constructor

        creates symbols
        """

    #
    #         self.logger.debug(f"===============Examples=================")
    #         if self.logger.getEffectiveLevel() <= logging.DEBUG:
    #             traceback.print_stack(file=sys.stdout)
    #
    #         self.args = {}  # args captured in here as name/value without "-" or "--"
    #         self.apply = []
    #         self.combine = None
    #         self.config = None  # holds configuration and global parameters (still being developed)
    #         self.current_file = None
    #         self.fileset = None
    #         # self.file_dict = {}  # possibly to be replaced by content_store.file_dict
    #         # self.content_store = ContentStore(self)  # will expose content_store.file_dict
    #         self.func_dict = {}
    #         self.result = None
    #         self.flag_dict = {}
    #         self.initialize_flags()
    #         self.wikidata_lookup = None
    #         self.wikipedia_lookup = None
    #         self.hit_counter = None
    #         # self.symbol_ini = SymbolIni(self)
    #         self.set_funcs()
    #         self.show_symbols = False
    #         self.ami_dictionary = None
    #         self.proj = None  # current project in searches
    #         self.current_ctree = None  # current ctree (may change during iteration
    #         self.cproject = None
    #         self.ami_logger = None
    #         self.outfile = None
    #         if self.show_symbols:
    #             pprint.pp(f"SYMBOLS\n {self.symbol_ini.symbols}")
    #
    #     def initialize_flags(self):
    #         """initialises flag_dict
    #         """
    #
    #         self.flag_dict = {
    #
    #             self.APPLY: None,
    #             self.CHECK_URLS: None,
    #             self.COMBINE: None,
    #             self.PRINT_SYMBOLS: False,
    #             self.RECURSE: True,
    #         }
    #
    #     def set_funcs(self):
    #         """initializes func_dict
    #         """
    #         # 1:1 methods
    #         # tuple of func+file_extnsion
    #         self.func_dict[self.XML2TXT] = (XmlLib.remove_all_tags, ".xml.txt")
    #         # self.func_dict[self.PDF2TXT] = (PdfReader.read_and_convert, ".pdf.txt")
    #         # self.func_dict[self.PDF2SVG] = (Pdf2SvgConverter.read_and_convert, ".pdf.svg")
    #         # self.func_dict[self.SVG2PAGE] = (Svg2PageConverter.read_and_convert, ".svg.xml")
    #         # self.func_dict[self.XML2HTML] = (Xml2HtmlConverter.read_and_convert, ".svg.html")
    #         # self.func_dict[self.TXT2SENT] = (TextUtil.split_into_sentences, ".sen.txt")
    #         # 1:n methods
    #
    def create_arg_parser(self):
        """creates adds the arguments for pyami commandline
        """

        def run_dict(self):
            print(f"run dict pyamix")

        def run_pdf(args):
            print(f"run pdf")

        # def run_project():
        #     print(f"run project {self.args}")

        version = self.version()
        if not sys.argv or len(sys.argv) == 0:
            # what does this do?
            sys.argv = [AmiLib.AmiLib]
        parser = argparse.ArgumentParser(
            description=f'pyamihtmlx: V{version} call with ONE of subcommands (HTML,PDF, IPCC, UNFCCC), e.g. pyamihtmlx IPCC --help'
        )

        parser.add_argument('-v', '--version', action="store_true",
                            help=f"show version {version}")
        parser.formatter_class = argparse.RawDescriptionHelpFormatter
        parser.description = textwrap.dedent(
            'pyamihtml: create, manipulate, use CProject \n'
            '----------------------------------------\n\n'
            'Pyamihtml is the largest collection of functionality in the AMI system.'
            'It contains executable code and libraries to manage complex documents.\n'
            'A key structure is the corpus (CProject directory) which contains a list of subdirectories '
            '(CTrees) which themselves contain many different document features (text, tables, images, graphics.\n'
            'Pyamihtml can create, fill, manipulate, transform many of the components including PDF, HTML, TXT, images, CSV.\n'
            '\n'
            'The subcommands:\n\n'
            '  HTML <options>      # create/edit HTML\n'
            '  IPCC <options>      # customised IPCC tools\n'
            '  PDF <options>       # convert PDF into HTML and images\n'
            '  UNFCCC <options>    # customised UNFCCC tools\n'
            '\n'
            'After installation, run \n'
            '  pyamihtmlx <subcommand> <options>\n'
            '\n'
            '\nExamples (# foo is a comment):\n'
            '  pyamihtmlx        # runs help\n'
            '  pyamihtmlx -h     # runs help\n'
            '  pyamihtmlx PDF -h # runs PDF help\n'
            '  pyamihtmlx PDF --infile foo.pdf --outdir bar/ # converts PDF to HTML\n'
            # '  pyamihtmlx PROJECT --project foodir/ # converts all PDF in foodir to CTrees\n'
            '  pyamihtmlx IPCC --pdf2html file/ # converts pdf file to html \n'
            '  pyamihtmlx UNFCCC --dir file/ # converts pdf files to html \n'
            '\n'
            '----------------------------------------\n\n'
        )

        # TODO should tests be run from this menu

        subparsers = parser.add_subparsers(help='subcommands', dest="command")
        for choice in subparsers.choices:
            print(f">>> {choice}")

        html_parser = HTMLArgs().make_sub_parser(subparsers)
        pdf_parser = PDFArgs().make_sub_parser(subparsers)

        amilib_parser = AmiLibArgs().make_sub_parser(subparsers)

        parser.epilog = "other entry points run as 'python -m pyamihtmlx.ami_dict args' also ami_pdf, ami_project"
        parser.epilog = """run:
        pyamihtmlx <subcommand> <args>
          where subcommand is in   {DICT,GUI,HTML,PDF,PROJECT} and args depend on subcommand
        """

        return parser

    def commandline(self, commandline: str) -> None:
        """runs a commandline as a single string
        """
        if not commandline:
            self.run_command(["--help"])
        else:
            arglist = commandline.split(" ")
            self.run_command(arglist)

    def run_commands(self, arglistlist):
        """runs a list of commands

        :param arglistlist:  A list of commands (which are usually lists)

        for each list element uses run_command
        This allows for setup, assertions, etc.

        typical example:
        self.run_commands
        """
        if arglistlist is not None and isinstance(arglistlist, list):
            for arglist in arglistlist:
                self.run_command(arglist)

    def run_command(self, args):
        """parses cmdline, runs command and outputs symbols

        :param args: either a string or a list of strings

        if args is a string we split it at spaces into a list of strings

        """
        if isinstance(args, str):
            args = args.strip()
            args = args.split(" ")

        print(f"command: {args}")
        self.logger.debug(f"********** raw arglist {args}")
        test_catch = False
        if test_catch:  # try to trap exception
            try:
                self.parse_and_run_args(args)
            except Exception as e:
                print(f"ERROR {e.args} from {args}")
                logging.error(f"\n============PARSE ERROR==({e.__cause__})======\n")
                return
            if self.is_flag_true(self.PRINT_SYMBOLS):
                self.symbol_ini.print_symbols()
        else:
            self.parse_and_run_args(args)

        return

    def parse_and_run_args(self, arglist, debug=False):
        """runs cmds and makes substitutions (${...} then runs workflow

        :param arglist:

        """
        # no args, create help
        if not arglist:
            self.logger.warning("No args, running --help")
            arglist = ["--help"]
        parser = self.create_arg_parser()
        self.args = self.make_substitutions_create_arg_tuples(arglist, parser, debug=debug)
        self.logger.debug("ARGS before substitution: " + str(self.args))
        # this may be redundant
        self.substitute_args()
        self.logger.debug(f"self.args {self.args}")
        self.add_single_str_to_list()
        self.logger.debug("ARGS after substitution: " + str(self.args))
        self.set_loglevel_from_args()
        self.run_arguments()

    #
    def substitute_args(self):
        """ iterates through self.args and makes subsitutions
        May duplicates
        """
        new_items = {}
        # self.logger.debug(f"SYMBOLS1 {self.symbol_ini.symbols}")
        for item in self.args.items():
            new_item = self.make_substitutions(item)
            self.logger.debug(f"++++++++{item} ==> {new_item}")
            new_items[new_item[0]] = new_item[1]
        self.args = new_items
        self.logger.debug(f"******** substituted ARGS {self.args}")

    #
    def add_single_str_to_list(self):
        """convert single strings to list of one string
        Not sure of what this is for
        """
        str_args = [self.DEBUG, self.EXAMPLES]
        for str_arg in str_args:
            self.logger.debug(f"key {str_arg}")
            self.replace_single_values_in_self_args_with_list(str_arg)
            self.logger.debug(f"args => {self.args}")

    #
    def run_arguments(self):
        """ parse and expland arguments then ru options for

        Currently:
        * examples
        * project
        * tests

        There will be more here

         """
        # print(f"RUN ARGUMENTS on {self} {self.args}")
        # path workflow
        self.wikipedia_lookup = WikidataLookup()
        self.logger.debug(f"commandline args {self.args}")
        subparser_type = self.args.get("command")
        logging.debug(f" COMMAND: {subparser_type} {self.args}")

        subparser_dict = {
            # "DICT": AmiDictArgs(),
            # "GUI":  GUIArgs(),
            "HTML": HTMLArgs(),
            "PDF": PDFArgs(),
        }
        abstract_args = subparser_dict.get(subparser_type)

        if abstract_args:
            abstract_args.parse_and_process1(self.args)
        else:
            self.run_core_mathods()

    #
    def run_core_mathods(self):
        logging.debug(f"run_core")
        # mainly obsolete
        if self.VERSION in self.args and self.args[self.VERSION] is not None:
            self.print_version()
        if self.FLAGS in self.args and self.args[self.FLAGS] is not None:
            self.add_flags()
        # if self.CONFIG in self.args and self.args[self.CONFIG] is not None:
        #     self.apply_config()
        # if self.EXAMPLES in self.args:
        #     example_args = self.args[self.EXAMPLES]
        #     if example_args is not None:
        #         self.logger.debug(f" examples args: {example_args}")
        #         Examples(self).run_examples(example_args)
        if self.COPY in self.args and not self.args[self.COPY] is None:
            self.logger.warning(f"COPY {self.args[self.COPY]}")
            self.copy_files()
        if self.PROJ in self.args:
            if self.SECT in self.args or self.GLOB in self.args or self.SPLIT in self.args:
                self.run_project_workflow()
        if self.VERSION in self.args:
            self.print_version()
        # elif TestFile.TEST in self.args:
        #     self.logger.warning(f"TEST in **args {self.args}")
        #     TestFile.run_arg_tests(self.args)
        # TODO linkup with test arguments

    def print_version(self):
        print(f"version {self.version()}")

    #
    def replace_single_values_in_self_args_with_list(self, key):
        """always returns list even for single arg
        e.g. turns "foo" into ["foo"]
        This is to avoid strings being interpreted as lists of characters
        I am sure there is a more pythonic way
        """
        # argsx = None
        if self.args is None:
            self.logger.warning(f"NULL self.args")
        elif key in self.args:
            argsx = self.args[key]
            if argsx is not None:
                if type(argsx) != list:
                    self.args[key] = [argsx]

    #
    def make_substitutions(self, item):
        """

        :param item:

        """
        # no change
        return item

        # old_val = item[1]
        # key = item[0]
        # # new_val = None
        # if True:
        #     new_val = old_val
        # elif old_val is None:
        #     new_val = None
        # elif isinstance(old_val, list) and len(old_val) == 1:  # single string in list
        #     # not sure of list, is often used when only one value
        #     val_item = old_val[0]
        #     new_val = self.symbol_ini.replace_symbols_in_arg(val_item)
        # elif isinstance(old_val, list):
        #     new_list = []
        #     for val_item in old_val:
        #         self.logger.debug(f"OLD SYM {val_item}")
        #         new_v = self.symbol_ini.replace_symbols_in_arg(val_item)
        #         self.logger.debug(f"NEW SYM {new_v}")
        #         new_list.append(new_v)
        #     self.logger.debug(f"UPDATED LIST ITEMS: {new_list}")
        #     new_val = new_list
        # elif isinstance(old_val, (int, bool, float, complex)):
        #     new_val = old_val
        # elif isinstance(old_val, str):
        #     if self.symbol_ini:
        #         new_val = self.symbol_ini.replace_symbols_in_arg(old_val)
        #         if "${" in new_val:
        #             raise ValueError(f"Unresolved reference : {new_val}")
        #             # print("=====================")
        #             # self.symbol_ini.print_symbols()
        #             # print("=====================")
        #             # new_val = self.symbol_ini.replace_symbols_in_arg(old_val)
        #         # else:
        #         #     new_val = old_val
        #         # new_items[key] = new_val
        # else:
        #     if type(old_val) is str:
        #         self.logger.error(f"substitution: {old_val} unknown arg type {type(old_val)}")
        #     new_val = old_val
        #
        # # special symbols such as "_proj"
        # if new_val:
        #     self.add_special_keys_to_symbols_ini(key, new_val)
        # return key, new_val

    # def get_symbol(self, symb):
    #     """gets symbol from pyami symbol table
    #
    #     """
    #     return self.symbol_ini.symbols.get(symb)
    #

    def make_substitutions_create_arg_tuples(self, arglist, parser, debug=False):
        """
        processes raw args to expand substitutions

        :param arglist:
        :param parser:
        :return: list of transformed arguments as 2-tuples
        """
        new_items = {}
        if arglist and len(arglist) > 0:
            parsed_args = self.parse_args_and_trap_errors(arglist, parser)
            if parsed_args is None:
                print(f"PARSED ARGS FAILS {arglist}")
                return new_items
            if parsed_args == self.SYSTEM_EXIT_OK:  # return code 0
                return new_items
            if str(parsed_args).startswith(self.SYSTEM_EXIT_FAIL):
                raise ValueError(f"bad command arguments {parsed_args} (see log output)")

            self.logger.debug(f"PARSED_ARGS {parsed_args}")
            if debug:
                print(f"parsed args {parsed_args}")

            try:
                arg_vars = vars(parsed_args)  # FAILS in Pycharm
            except Exception as e:
                print(
                    f"argparse and vars() fails, see \n BUG in Pycharm/Pandas see https://stackoverflow.com/questions/75453995/pandas-plot-vars-argument-must-have-dict-attribute\n try fudge")
                # Namespace(command=None, version=True)
                match = re.match("Namespace\\((?P<argvals>[^)]*)\\)", parsed_args)
                if match:
                    arglistx = match.groupdict("argvals")
                    arg_vars = arglistx.split(",\\s*")

            for item in arg_vars.items():  # BUG in Pycharm/Pandas see https://stackoverflow.com/questions/75453995/pandas-plot-vars-argument-must-have-dict-attribute
                new_item = self.make_substitutions(item)
                new_items[new_item[0]] = new_item[1]
        return new_items

    #
    def parse_args_and_trap_errors(self, arglist, parser):
        """run argparse parser.parse_args and try to trap serious errors
        --help calls SystemExit (we trap and return None)"""
        try:
            parsed_args = parser.parse_args(arglist)
        except SystemExit as se:  # exit codes
            if str(se) == '0':
                parsed_args = self.SYSTEM_EXIT_OK
            else:
                parsed_args = self.SYSTEM_EXIT_FAIL + str(se)
        except Exception as e:
            parsed_args = None
            self.logger.error(f"Cannot parse {arglist} , {e}")
        return parsed_args

    #
    # def add_special_keys_to_symbols_ini(self, key, value):
    #     """
    #
    #     :param key:
    #     :param value:
    #
    #     """
    #     if key in self.SPECIAL_SYMBOLS:
    #         self.symbol_ini.symbols[key] = value
    #         self.logger.warning(f"added reserved symbol {key} => {value}")
    #
    def set_loglevel_from_args(self):
        """ """
        levels = {
            "debug": logging.DEBUG,
            "info": logging.INFO,
            "warning": logging.WARNING,
            "error": logging.ERROR,
        }

        if self.LOGLEVEL in self.args:
            loglevel = self.args[self.LOGLEVEL]
            self.logger.info(f"loglevel {loglevel}")
            if loglevel is not None:
                loglevel = str(loglevel)
            if loglevel is not None and loglevel.lower() in levels:
                level = levels[loglevel.lower()]
                self.logger.loglevel = level

    #
    # def run_project_workflow(self):
    #     """ run when PROJ is set"""
    #     self.logger.debug(f"ARGS {self.args}")
    #     if not self.args:
    #         self.logger.error("no args given; try --examples or --help")
    #         return
    #     if self.args[self.DEBUG]:
    #         self.run_debug()
    #     if self.args[self.PROJ]:
    #         self.hit_counter = Counter()
    #         self.run_proj()
    #         self.logger.debug(f"hit counter: {self.hit_counter}")
    #     # if self.args[self.TEST]:
    #     #     TestFile.run_arg_tests(self.args)
    #     # TODO fix tests
    #     # else:
    #     #     self.logger.error("{self.args} requires --proj or --test")
    #     return
    #
    # def run_debug(self):
    #     for arg in self.args[self.DEBUG]:
    #         if arg == self.SYMBOLS:
    #             if self.is_flag_true(self.PRINT_SYMBOLS):
    #                 self.symbol_ini.print_symbols()
    #         else:
    #             self.logger.warning(f"unknown arg {arg} in  debug: ")
    #
    # def run_proj(self):
    #     """ project-related commands"""
    #     print(f"run_proj obsolete")
    #     return
    #     # self.proj = self.args[self.PROJ]
    #     # if not self.proj:
    #     #     self.logger.error(f"--proj must be given")
    #     #     return
    #     # self.cproject = CProject(self.proj)
    #     #
    #     # self.logger.debug(f"{Util.basename(__file__)} proj: {self.proj}")
    #     # # if self.args[self.CONFIG]:
    #     # #     self.apply_config()
    #     # if self.args[self.DELETE]:
    #     #     self.delete_files()
    #     # if self.args[self.GLOB]:
    #     #     self.glob_files()
    #     # if self.args[self.SPLIT]:
    #     #     self.split(self.args.get(self.SPLIT))
    #     # if self.args[self.APPLY]:
    #     #     apply_type = self.args.get(self.APPLY)
    #     #     converter = Converters.get_converter(apply_type)
    #     #     if not converter:
    #     #         raise ValueError(f"cannot find converter for {apply_type}")
    #     #     previous = False
    #     #     if previous:
    #     #         self.add_ctree_filenames_to_content_store(apply_type)
    #     #         # print(f"content_store {self.content_store.file_dict}")
    #     #         self.apply_func(apply_type)
    #     #     else:
    #     #         converter = Converters.get_converter(apply_type)()
    #     #         print(f"{type(converter)}")
    #     #         converter.iterate_cproject(cproject=self.proj)
    #     #
    #     # if self.args[self.FILTER]:
    #     #     self.filter_file()
    #     # if self.args[self.COMBINE]:
    #     #     self.combine_files_to_object()
    #     # if self.args[self.OUTFILE]:
    #     #     self.write_output()
    #     # if self.args[self.ASSERT]:
    #     #     self.run_assertions()
    #
    # def copy_files(self):
    #     """copies path or directory
    #
    #     copies a path or complete directory
    #
    #     Args:
    #     Exceptions:
    #
    #     """
    #     # self.logger.warning(f"NOT IMPLERMENTED")
    #     # return
    #     self.replace_single_values_in_self_args_with_list(self.COPY)
    #     argsx = self.args[self.COPY]
    #     if len(argsx) < 2:
    #         raise TypeError("copy needs >= 2 args")
    #
    #     src = argsx[0]
    #     src_path = Path(src)
    #     if not src_path.exists():
    #         raise FileNotFoundError(
    #             str(src_path), "src must exist as path or directory")
    #
    #     dest = argsx[1]
    #     dest_path = Path(dest)
    #
    #     overwrite = len(argsx) > 2 and argsx[2] == "overwrite"
    #     FileLib.copy_file_or_directory(dest_path, src_path, overwrite)
    #
    # def add_flags(self):
    #     """uses values from '--flag [value]' arguments """
    #     flags = self.args[self.FLAGS]
    #     if type(flags) is str:
    #         flags = [flags]
    #     for flag in flags:
    #         nv = Util.create_name_value(flag)
    #         if nv is None or len(nv) != 2:
    #             raise ValueError(f"bad value {flag}")
    #         if self.flag_dict.get(nv[0]) is None:
    #             self.logger.warning(f"adding new flag {nv}")
    #         self.flag_dict[nv[0]] = nv[1]
    #
    # def apply_config(self):
    #     config_args = self.args[self.CONFIG]
    #     if type(config_args) == str:
    #         config_args = [config_args]
    #     self.logger.debug(f" type {type(config_args)}")
    #     for config_arg in config_args:
    #         if config_arg == self.SYMBOLS:
    #             if self.is_flag_true(self.PRINT_SYMBOLS):
    #                 self.symbol_ini.print_symbols()
    #         else:
    #             self.logger.warning(f"processing INI NYI: {config_arg}")
    #
    # def delete_files(self):
    #     """deletes files in glob
    #
    #     requires proj to be set
    #     """
    #     if self.proj is None or self.proj == "":
    #         self.logger.error(f"delete requires --proj; ignored")
    #         return
    #     globs = self.args[self.DELETE]
    #     for globx in globs:
    #         self.delete_glob(globx)
    #
    # def delete_glob(self, glob_exp):
    #     """deletes globbed files
    #     requires self.proj as root of globbing
    #
    #     Args:
    #         glob_exp (str): glob expression
    #
    #
    #     """
    #     if ".." in glob_exp or glob_exp.endswith("*"):
    #         self.logger.error(f"glob {glob_exp} cannot contain .. or end in *")
    #         return
    #     full_glob = self.proj + "/" + glob_exp
    #     self.logger.warning(f"delete: {full_glob}")
    #     glob_recurse = True  # change this later
    #     globs = FileLib.posix_glob(full_glob, recursive=glob_recurse)
    #     if globs is not None:
    #         files = {file: None for file in globs}
    #         self.logger.warning(f"deleting {len(files)} files ")
    #         for f in files:
    #             p = Path(f)
    #             if p.is_dir():
    #                 self.logger.warning(f"Cannot yet delete directories {p}")
    #             else:
    #                 p.unlink()
    #
    # def glob_files_posix(self):
    #     glob_recurse = self.is_flag_true(self.RECURSE)
    #     glob_ = self.args[self.GLOB]
    #     self.logger.info(f"glob: {glob_}")
    #     # create dictionary wiuth empty values?
    #     files = [file for file in FileLib.posix_glob(glob_, recursive=glob_recurse)]
    #     self.content_store.add_files(files)
    #
    #     self.logger.info(f"glob path count {len(self.content_store.keys())}")
    #
    # def split(self, typex):
    #     """ split fulltext.xml into sections"""
    #
    #     # file_keys = self.content_store.get_file_keys()
    #     file_keys = self.content_store.keys()
    #     for file in file_keys:
    #         suffix = FileLib.get_suffix(file)
    #         if ".xml" == suffix and typex == self.XML2SECT:
    #             print(f"make_xml_sections onsolete")
    #             # self.make_xml_sections(file)
    #             # force = False
    #             # force = True
    #             # outdir = Path(Path(file).parent, "sections")
    #             # AMIAbsSection.make_xml_sections(file, str(outdir), force)
    #         elif ".txt" == suffix or typex == self.TXT2PARA:
    #             print(f"TXT2PARA obsolete")
    #             # self.make_text_sections(file)
    #         else:
    #             self.logger.warning(f"no match for suffix: {suffix}")
    #
    # @classmethod
    # def make_xml_sections(cls, file):
    #     xml_libx = XmlLib()
    #     xml_libx.logger.setLevel(logging.DEBUG)
    #     xml_libx.read(file)
    #     xml_libx.make_sections("sections", )
    #
    # def make_text_sections(self, file):
    #     print(f"make_test_sections obsolete")
    #     # sections = []
    #     # with open(file, "r", encoding="utf-8") as f:
    #     #     text = f.read()
    #     #     sections = TextUtil.split_at_empty_newline(text)
    #     #     self.store_or_write_data(file, sections, )
    #
    # def apply_func(self, apply_type):
    #     """ """
    #     #
    #     self.read_file_content()
    #     if apply_type:
    #         self.logger.info(f"apply {apply_type}")
    #         func_tuple = self.func_dict[apply_type]
    #         if func_tuple is None:
    #             self.logger.error(f"Cannot find func for {apply_type}")
    #         else:
    #             # apply data is stored in self.file_dict
    #             self.logger.debug(f"running {func_tuple} {apply_type}")
    #             self.apply_to_file_content(func_tuple, apply_type)
    #
    #     return
    #
    # def normalize(self, unistr):
    #     import unicodedata
    #     self.logger.error("NYI")
    #     unicodedata.normalize('NFKC', unistr)
    #     pass
    #
    # def filter_file(self):
    #     filter_expr = self.args[self.FILTER]
    #     self.logger.warning(f"filter: {filter_expr}")
    #
    #     files = set()
    #     # record hits
    #     file_keys = self.content_store.keys()
    #     for file in file_keys:
    #         filter_true = self.apply_filter(file, filter_expr)
    #         if filter_true:
    #             files.add(file)
    #     # delete hits from dict
    #     for file in files:
    #         if file in self.content_store.keys():
    #             del self.content_store.file_dict[file]
    #
    # def apply_filter(self, file, filter_expr):
    #     found = False
    #     if not filter_expr:
    #         self.logger.error(f"No filter expression")
    #         return found
    #     with open(file, "r", encoding="utf-8") as f:
    #         content = f.read()
    #     hits = []
    #     if isinstance(filter_expr, list):
    #         # and'ed at present
    #         for f_expr in filter_expr:
    #             hits = self.apply_filter_expr(content, file, f_expr, hits)
    #     else:
    #         hits = self.apply_filter_expr(content, file, filter_expr, hits)
    #
    #     return hits
    #
    # def apply_filter_expr(self, content, file, filter_expr, hit_list):
    #     """ applies filters to hit list, usually AND"""
    #     self.logger.debug(f"filter_expr {filter_expr}")
    #     filter_expr = filter_expr.strip()
    #     filter_value = self.extract_command_value(filter_expr)
    #     if filter_value is None:
    #         self.logger.error(f"bad filter_expr {filter_expr}")
    #         return hit_list
    #     filterx = filter_value[0]
    #     value = filter_value[1]
    #     value = self.symbol_ini.replace_symbols_in_arg(value)
    #     if value is None:
    #         self.logger.warning(f"null value in filter {filterx}")
    #         return None
    #     if filterx == self.CONTAINS and file.endswith(".txt"):
    #         if value in content:
    #             hit_list.append(value)
    #
    #     elif filterx == self.DICTIONARY and file.endswith(".xml"):
    #         print(f"dictionary nlot supported yet")
    #         # hits = self.apply_dictionary(hit_list, value)
    #         # if len(hits) > 0:
    #         #     self.logger.debug(f"xpath {type(hits)} {hits}")
    #         #     hit_list.extend(hits)
    #
    #     elif filterx == self.LOOKUP:
    #         self.logger.debug(f"LOOKUP VALUE {value}")
    #         hits = self.apply_lookup(hit_list, value)
    #         if hits:
    #             hit_list = hits
    #     elif filterx == self.REGEX:
    #         hits = self.apply_regex(hit_list, value)
    #         if hits:
    #             self.logger.debug(f"regex hits {hits}")
    #             hit_list = hits
    #     elif filterx == self.WIKIDATA_SPARQL:
    #         hits = self.apply_wikidata_sparql(hit_list, value)
    #         if hits:
    #             self.logger.debug(f"wikidata_sparql hits {hits}")
    #             hit_list = hits
    #
    #     elif filterx == self.XPATH and file.endswith(".xml"):
    #         tree = etree.parse(file)
    #         hits = [h.strip() for h in tree.xpath(value)]
    #         if len(hits) > 0:
    #             self.logger.warning(f"xpath {type(hits)} {hits}")
    #             hit_list.extend(hits)
    #
    #     self.logger.debug(f"hit list {hit_list}")
    #     if hit_list:
    #         self.logger.info(f"non-zero list {hit_list}")
    #     return hit_list
    #
    # @classmethod
    # def extract_command_value(cls, command_expr):
    #     """split command(value) into tuple
    #
    #     value may have nested commands.
    #
    #     :returns: tuple of command, value
    #     """
    #     if command_expr is None:
    #         return None
    #     bits = command_expr.split("(", 1)
    #     cls.logger.debug(f"BITS {bits}")
    #
    #     return (bits[0], bits[1][:-1]) if len(bits) > 1 and bits[1].endswith(")") else None
    #
    # # def apply_dictionary(self, hits, name):
    # #
    # #     dictionary_file = self.get_symbol(name)
    # #     if dictionary_file is None:
    # #         dictionary_file = name
    # #     self.ami_dictionary = AmiDictionary.read_dictionary(
    # #         file=dictionary_file)
    # #     new_hits = []
    # #     if self.ami_dictionary is not None:
    # #         for hit in hits:
    # #             entry = self.ami_dictionary.get_lxml_entry(hit.lower())
    # #             if entry is not None:
    # #                 new_hits.append(hit)
    # #                 self.hit_counter[hit] += 1
    # #
    # #     # return [hit for hit in hits if re.match(regex, hit)]
    # #     return new_hits
    #
    # @classmethod
    # def apply_regex(cls, hits, regex):
    #     return [hit for hit in hits if re.match(regex, hit)]
    #
    # def add_ctree_filenames_to_content_store(self, apply_type):
    #     print(f"add_ctree_filenames_to_content_store obsolete")
    #     # # files = []
    #     # ctree_list = self.cproject.get_ctrees()
    #     # # TODO add this functionality to enums
    #     # subdir_name = None
    #     # if apply_type == self.SVG2PAGE:
    #     #     subdir_name = CTree.SVG_DIR
    #     #     subdir_ext = "svg"
    #     #
    #     # subdirs = []
    #     # subfiles = []
    #     # for ctree in ctree_list:
    #     #     sd = ctree.get_existing_reserved_directory(subdir_name)
    #     #     subdir = CSubDir(sd)
    #     #     subdirs.append(subdir)
    #     #     globx = "*.svg"
    #     #     sf = subdir.get_descendants(globx)
    #     #     subfiles.extend(sf)
    #     #
    #     # print(f"subdirs {len(subdirs)} subfiles {len(subfiles)}")
    #     # self.content_store.add_files(subfiles)
    #
    # # def get_search_string(self, filter_expr, search_method):
    # #     return filter_expr[len(search_method) + 1:-1]
    # #
    # def read_file_content(self, to_str=True):
    #     """read path content as bytes into file_dict
    #
    #     :to_str: if true convert content to strings
    #
    #     :param to_str:  (Default value = True)
    #
    #     """
    #     self.ami_logger = AmiLogger(self.logger, initial=10, routine=100)
    #     keys = self.content_store.keys()
    #     print(f"keys {len(keys)}")
    #     for file in keys:
    #         filestr = str(file)
    #         self.logger.info(f"reading... {file}")
    #         if filestr.endswith(".xml"):
    #             self.read_string_content_to_dict(file, to_str)
    #         elif filestr.endswith(".svg"):
    #             # self.save_file_name_to_dict(file)
    #             self.read_string_content_to_dict(file, to_str)
    #         elif filestr.endswith(".pdf"):
    #             self.save_file_name_to_dict(file)
    #         elif filestr.endswith(".png"):
    #             self.read_binary_content_to_dict(file)
    #         elif filestr.endswith(".txt"):
    #             self.read_string_content_to_dict(file, to_str=False)
    #         else:
    #             self.logger.warning(f"cannot read path into string {file}")
    #
    # def apply_lookup(self, hits, value):
    #     self.logger.debug(f"LOOKUP: {hits} {value}")
    #     for hit in hits:
    #         if self.get_dictionary(value) is not None:
    #             dictionary = self.get_dictionary(value)
    #             self.logger.warning("USE DICTIONARY: NYI" + str(value) + str(dictionary))
    #         elif value == 'wikidata':
    #             qnumber = self.wikipedia_lookup.lookup_wikidata(hit)
    #             self.ami_logger.info(f"qnumber {qnumber}")
    #         else:
    #             self.logger.error(f"cannot parse lookup: {value}")
    #
    # def apply_wikidata_sparql(self, hit_list):
    #     if hit_list:
    #         self.logger.warning(f"wikidata input {hit_list}")
    #     return hit_list
    #
    # def get_dictionary(self, value):
    #     dictionary = None
    #     command_value = self.extract_command_value(value)
    #     if command_value is not None and command_value[0] == "dictionary":
    #         dictionary = command_value[1]
    #     return dictionary
    #
    # def read_string_content_to_dict(self, file, to_str):
    #     """reads path into string
    #     Can process bytes to string
    #
    #     DO WE NEED TO STORE THIS?
    #     """
    #     data = None
    #     self.logger.info(f"reading string content from {file}")
    #     with open(file, "r", encoding="utf-8") as f:
    #         try:
    #             data = f.read()
    #             if to_str and isinstance(data, bytes):
    #                 data = data.decode("utf-8")
    #             self.store_or_write_data(file, data)
    #             # self.file_dict[path] = data
    #         except UnicodeDecodeError as ude:
    #             self.logger.error(f"skipped decoding error {ude}")
    #     return data
    #
    # def save_file_name_to_dict(self, file):
    #     self.store_or_write_data(file, file)
    #     # self.file_dict[path] = path
    #
    # def read_binary_content_to_dict(self, file):
    #     with open(file, "rb", ) as f:
    #         try:
    #             data = f.read()
    #             self.store_or_write_data(file, data)
    #             # self.file_dict[path] = data
    #         except Exception as e:
    #             self.logger.error(f"skipped reading error {e}")
    #
    # def apply_to_file_content(self, func_tuple, apply_type):
    #     """applies func to all string content in file_dict
    #
    #     :param func_tuple:
    #     :param apply_type:
    #
    #     """
    #     self.logger.warning(f"func_tuple {len(func_tuple)}: {func_tuple[0]} => {func_tuple[1]}")
    #
    #     open_flag = self.get_open_type(apply_type)
    #     for file in self.content_store.keys():
    #         self.logger.warning(f"converting {file}")
    #         data = self.content_store.get_file_contents(file)
    #         if not data:
    #             encoding = "utf-8" if open_flag == "r" else None
    #             with open(file, open_flag, encoding=encoding) as f:
    #                 data = f.read()
    #         print(f"file {file} data {len(data)}")
    #         self.logger.debug(f"path: {file} => {func_tuple[0]}")
    #         new_file = self.create_file_name(file, func_tuple[1])
    #         print(f"new file: {new_file}")
    #
    #         try:
    #             new_data = func_tuple[0](data)
    #             print(f"new {len(new_data)}")
    #         except Exception as err:
    #             print(f"cannot convert {file} because {err} skipped")
    #             return
    #
    #         try:
    #             self.store_or_write_data(file, new_data, new_file)
    #         except Exception as err:
    #             print(f"cannot save {new_file} because {err} skipped")
    #             return
    #     return
    #
    # def get_open_type(self, apply_type):
    #     """ gets 'rb' for binary files or 'r' for text"""
    #     open_type = "rb"
    #     if str(apply_type) in [self.SVG2PAGE, self.XML2TXT, self.XML2HTML, self.TXT2SENT]:
    #         open_type = "r"
    #     return open_type
    #
    # # needs fixing
    # @classmethod
    # def create_file_name(cls, file, extension):
    #     pathname = Path(file)
    #     return str(pathname.with_suffix(extension))
    #
    # def store_or_write_data(self, file, data, new_file=None) -> None:
    #     """store or write data to disk"""
    #     if file in self.content_store.file_dict:
    #         old_data = self.content_store.file_dict[file]
    #         if old_data is not None and old_data != data:
    #             self.logger.warning(
    #                 f"===============================\n"
    #                 f"=========OVERWRITING data for {file}\n"
    #                 f"{self.content_store.file_dict[file]} \n========WITH======\n"
    #                 f"{data}")
    #             if new_file is not None:
    #                 self.logger.warning(f"WROTE: {new_file}")
    #                 with open(new_file, "w", encoding="utf-8") as f:
    #                     f.write(data)
    #                 self.content_store.file_dict[file] = new_file
    #
    #     # save data old-style
    #     self.content_store.file_dict[file] = data
    #
    # def combine_files_to_object(self):
    #     """ """
    #     methods = self.args.get(self.COMBINE)
    #     if methods and methods == self.CONCAT_STR:
    #         self.result = "\n".join(self.content_store.file_dict.values())
    #         self.logger.warning(f"combine {self.result}")
    #
    # def write_output(self):
    #     """ """
    #     self.outfile = self.args[self.OUTFILE]
    #     if self.result:  # single output
    #         self.write_single_result()
    #
    #     if self.content_store.file_dict:
    #         self.write_multiple_results()
    #
    # def write_multiple_results(self):
    #     for file in self.content_store.file_dict:
    #         data = self.content_store.file_dict[file]
    #         if data is None:
    #             self.logger.debug(f"data is NONE")
    #             continue
    #         print(f"Valid data")
    #         parent = FileLib.get_parent_dir(file)
    #         new_outfile = os.path.join(parent, self.outfile)
    #         try:
    #             FileLib.force_mkparent(new_outfile)
    #             self.logger.warning(f"writing path {new_outfile}")
    #             with open(new_outfile, "w", encoding="utf-8") as f:
    #                 self.logger.warning(f"wrote results {new_outfile}")
    #                 f.write(f"{str(data)}")
    #         except Exception as e:
    #             self.logger.warning(f"cannot write because {e}")
    #
    # def write_single_result(self):
    #     FileLib.force_write(self.outfile, self.result, overwrite=True)
    #     self.logger.warning(f"wrote results {self.outfile}")
    #
    # def run_assertions(self):
    #     """ """
    #     print(f"assertiions not run")
    #     return
    #     # assertions = self.args.get(self.ASSERT)
    #     # if assertions is not None:
    #     #     self.parser = DSLParser()
    #     #     if isinstance(assertions, str):
    #     #         assertions = [assertions]
    #     #     for assertion in assertions:
    #     #         self.parser.parse_and_run(assertion)
    #
    # def is_flag_true(self, flag):
    #     """is flag set in flag_dict
    #     values can be None, "None", "True", True, "False", False or strings
    #     Messay
    #     :param flag:
    #
    #     """
    #     val = self.flag_dict.get(flag)
    #     if not val or val == "None" or val == "False":
    #         return False
    #     if val:
    #         return True
    #     try:
    #         return ast.literal_eval(val)
    #     except Exception:
    #         return False

    def version(self):
        """
        reads setup.py and extracts line of form version='0.0.29'
        This is still a mess. We need to set the version once somewhere.
        """

        version = '0.0.1a1'  # 2024-03-27
        version = '0.0.1a3'  # 2024-03-27
        version = '0.0.1'  # 2024-04-03
        version = '0.0.2'  # 2024-04-04
        version = '0.0.3'  # 2024-04-19
        # had to revert here I think
        version = '0.0.6'  # 2024-04-19

        # logging.warn(f"VERSION {version}")
        return version


class AmiLibArgs(AbstractArgs):
    """Parse args to analyze, edit and annotate HTML"""

    def __init__(self):
        """arg_dict is set to default"""
        super().__init__()
        self.dictfile = None
        self.inpath = None
        self.outpath = None
        self.outstem = None
        self.outdir = None
        self.arg_dict = None
        self.subparser_arg = "HTML"

    def add_arguments(self):
        if self.parser is None:
            self.parser = argparse.ArgumentParser()
        """adds arguments to a parser or subparser"""
        self.parser.description = 'HTML editing analysing annotation'
        self.parser.add_argument(f"--foo", action="store_true",
                                 help="annotate HTML file with dictionary")
        # self.parser.add_argument(f"--{COLOR}", type=str, nargs=1,
        #                          help="colour for annotation")
        # self.parser.add_argument(f"--{DICT}", type=str, nargs=1,
        #                          help="dictionary for annotation")
        # self.parser.add_argument(f"--{INPATH}", type=str, nargs=1,
        #                          help="input html file")
        # self.parser.add_argument(f"--{OUTPATH}", type=str, nargs=1,
        #                          help="output html file")
        # self.parser.add_argument(f"--{OUTDIR}", type=str, nargs=1,
        #                          help="output directory")
        self.parser.epilog = "==============="

    """python -m pyamihtmlx.pyamix HTML --annotate 
     --dict /Users/pm286/projects/semanticClimate/ipcc/ar6/wg3/Chapter02/dict/emissions_abbreviations.xml
     --inpath /Users/pm286/projects/semanticClimate/ipcc/ar6/wg3/Chapter02/fulltext.html
     --outpsth /Users/pm286/projects/semanticClimate/ipcc/ar6/wg3/Chapter02/annotated/fulltext_emissions.html
     --color pink
     """

    # class AmiDictArgs:
    def process_args(self):
        """runs parsed args
        :return:
        """

        if not self.arg_dict:
            logging.warning(f"no arg_dict given, no action")
            return

        self.foo = self.arg_dict.get("foo")
        # self.color = self.arg_dict.get(COLOR)
        # self.dictfile = self.arg_dict.get(DICT)
        # self.inpath = self.arg_dict.get(INPATH)
        # self.outdir = self.arg_dict.get(OUTDIR)
        # self.outpath = self.arg_dict.get(OUTPATH)

        if self.foo:
            self.make_foo()

    # class AmiDictArgs:

    @classmethod
    def create_default_arg_dict(cls):
        """returns a new COPY of the default dictionary"""
        arg_dict = dict()
        arg_dict["dict"] = None
        return arg_dict

    @property
    def module_stem(self):
        """name of module"""
        return Path(__file__).stem

    def annotate_with_dict(self):
        """uses dictionary to annotate words and phrases in HTML file"""
        self.logger.warning("Dictionaries not supported")
        return

        if not self.dictfile:
            logging.error(f"no dictionary given")
            return
        if not self.inpath:
            logging.error(f"no input file to annotate given")
            return
        if not self.outpath:
            logging.error(f"no output file given")
            return
        if not self.outdir:
            self.outdir = Path(self.outpath).parent
        self.outdir = Path(self.outdir)
        self.outdir.mkdir(exist_ok=True)

        self.ami_dict = AmiDictionary.create_from_xml_file(self.dictfile)
        self.ami_dict.markup_html_from_dictionary(self.inpath, self.outpath, self.color)

    def make_foo(self):
        pass


class Converter(Enum):

    def __init__(self, converter_class, intype, outtype, indir=".", outdir="."):
        self.intype = intype
        self.indir = indir
        self.outtype = outtype
        self.outdir = outdir

    # PDF2SVG = (Pdf2SvgConverter, "pdf", "svg", ".", "svg")
    # PDF2TXT = (PdfReader, Filetype.F_PDF, Filetype.F_TXT, ".", ".")
    # XML2HTML = (Xml2HtmlConverter, "pdf", "html", ".", ".")
    # XML2TXT = (Xml2TxtConverter, "xml", "txt", ".", ".")
    # SVG2PAGE = (Svg2PageConverter, "svg", "html", "svg", "page")
    # PAGE2SECT = (Page2SectConverter, "html", "html", "page", "sect")
    # TXT2SENT = (Txt2SentSplitter, Filetype.F_TXT, Filetype.F_TXT, ".", "sent")


def main():
    # make_cmd()
    """ main entry point for cmdline

    """
    print(f"PYAMI")
    run_tests = False  # needs re-implementing

    run_commands = True
    #    run_commands = False

    #    run_tests = True

    AmiLib.logger.debug(
        f"\n============== running pyami main ===============\n{sys.argv[1:]}")
    amix = AmiLib()
    print(f"***** PYAMI VERSION {amix.version()} *****")
    # this needs commandline
    if run_commands:
        amix.run_command(sys.argv[1:])


if __name__ == "__main__":
    main()
